/*
 * Copyright (c) 1997, 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <windows.h>
#include <GL\gl.h>
#include "handy.h"

static void (*_SendToVideo)();
static void (*_GetInputFields)();
static void (*_GetInputArray)();

/* Extension function pointers */
static PFNGLBLENDCOLOREXTPROC     pglBlendColor;
static PFNGLBLENDEQUATIONEXTPROC  pglBlendEquation;


__declspec(dllexport)
void LoadGLExtensions (void)
{
    pglBlendColor = (PFNGLBLENDCOLOREXTPROC) wglGetProcAddress("glBlendColorEXT");
    pglBlendEquation = (PFNGLBLENDEQUATIONEXTPROC)wglGetProcAddress("glBlendEquationEXT");
}

__declspec(dllexport)
void SetVideoSendRoutine(void (*f)())
{
    _SendToVideo = f;
}

__declspec(dllexport)
void SetVideoGetFieldsRoutine(void (*f)())
{
    _GetInputFields = f;
}

__declspec(dllexport)
void SetVideoGetArrayRoutine(void (*f)())
{
    _GetInputArray = f;
}

__declspec(dllexport)
void do_single(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
               int tex_name, int *playing, int ec)
{
    int count = 0;
    char *sourceBuffer1, *sourceBuffer2;

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);

    while (*playing != ec) {
        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer1);
        count++;
        (*_SendToVideo) (count);
    }

}

__declspec(dllexport)
void do_blend(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
              int tex_name, int *playing, int ec)
{
    int i, val;
    long x, y;
    int count = 0;
    float bf = 0;
    float bf_inc;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    bf_inc = .01;
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_CONSTANT_ALPHA_EXT,GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
    (*pglBlendEquation)(GL_FUNC_ADD_EXT);
    glLoadIdentity();

    while (*playing != ec) {
        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glDisable(GL_BLEND);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer1);

        glEnable(GL_BLEND);
        (*pglBlendColor)(0.0, 0.0, 0.0, bf);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        count++;
        (*_SendToVideo) (count);
    
        bf += bf_inc;
        if ((bf > .995) || (bf < 0.005) ) {
            bf_inc = -bf_inc;
        } 
    }
}

__declspec(dllexport)
void do_wipe(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
             int tex_name, int *playing, int ec)
{
    /* This routine uses stencil planes to draw a wipe (thanks to the idea from
     * true@boston.sgi.com).  It could be modified easily to draw a large variety
     * of standard SMPTE wipes).
     */
    int i, val;
    long x, y;
    int count = 0;
    float bf = -0.9;
    float bf_inc = .01;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glLoadIdentity();
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    bf_inc = .01;

    while (*playing != ec) {
        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        /* Draw the background - no stenciling yet */
	    glDisable(GL_STENCIL_TEST);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer1);

        /* Clear the stencil plane */
        glClearStencil(0x0);
        glClear(GL_STENCIL_BUFFER_BIT);

        /* The selected stencil function will always fail (since
         * the stencil plane is filled with 0's it will never equal
         * 1).  Whenever it fails, have glStencilOp load a 1 into the
         * the stencil buffer.  This loads a 1 into the stencil buffer wherever
         * the rectangle is drawn (and 0 otherwise).
         */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_EQUAL, 0x1, 0x1);
        glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
        glRectf(-1.0, -1.0, bf, 1.0);

        /* The stencil function is still "equals 1", so wherever the rectangle
         * has been drawn, the glDrawPixels will show through.
         */ 
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 0.9) || (bf < -0.9) ) {
            bf_inc = -bf_inc;
        }
    }
    glDisable(GL_STENCIL_TEST);
}

__declspec(dllexport)
void do_pip(int uSize, int vSize, int tex_width, int tex_height, int tex_format, int tex_name, 
            int *playing, int ec)
{
    int i, val;
    long x, y;
    int count = 0;
    float bf = 0;
    float bf_inc;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);
    bf_inc = .01;
    glLoadIdentity();
    glScalef(0.5, 0.5, 0.5);

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glDisable(GL_TEXTURE_2D);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);
        glEnable(GL_TEXTURE_2D);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer1);
        glTranslatef(bf_inc, 0.0, 0.0);
        glCallList(1);

        /* Draw a border around the pip.	*/
        glDisable(GL_TEXTURE_2D);
        glColor4f(0.0, 0.0, 1.0, 1.0);

        glRectf(-1.2, -1.2, 1.2, -.98);  /* bottom */
        glRectf(-1.2,  1.2, 1.2,  .98);  /* top    */

        glRectf(-1.2, -1.2, -.98, 1.2);  /* left   */
        glRectf( 1.2, -1.2,  .98, 1.2);  /* right  */

        glColor4f(1.0, 1.0, 1.0, 1.0);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 1.995) || (bf < -1.995) ) {
            bf_inc = -bf_inc;
        }
    }
    glColor4f(1.0, 1.0, 1.0, 1.0);
}

__declspec(dllexport)
void do_cross(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
              int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    long x, y;
    float bf = 0;
    float bf_inc;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);
    bf_inc = .01;
    glLoadIdentity();

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glDisable(GL_TEXTURE_2D);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        glEnable(GL_TEXTURE_2D);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, GL_UNSIGNED_BYTE, 
                           sourceBuffer1);

        glLoadIdentity();
        glTranslatef(-bf, -bf, 0.0);
        glCallList(2);

        glLoadIdentity();
        glTranslatef(bf, -bf, 0.0);
        glCallList(3);

        glLoadIdentity();
        glTranslatef(bf, bf, 0.0);
        glCallList(4);

        glLoadIdentity();
        glTranslatef(-bf, bf, 0.0);
        glCallList(5);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 1.0) || (bf < -1.0) ) {
            bf_inc = -bf_inc;
        }
    }

}

__declspec(dllexport)
void do_boxblend(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                 int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    long x, y;
    float bf = 0;
    float bf_inc;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);
    glEnable(GL_BLEND);
    glBlendFunc(GL_CONSTANT_ALPHA_EXT,GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
    (*pglBlendEquation)(GL_FUNC_ADD_EXT);

    bf_inc = .01;
    glLoadIdentity();

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glDisable(GL_BLEND);
        glDisable(GL_TEXTURE_2D);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        glEnable(GL_TEXTURE_2D);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer1);

        glEnable(GL_BLEND);

        (*pglBlendColor)(0.0, 0.0, 0.0, bf);
        glCallList(2);

        (*pglBlendColor)(0.0, 0.0, 0.0, 1.0 - bf);
        glCallList(3);

        (*pglBlendColor)(0.0, 0.0, 0.0, bf);
        glCallList(4);

        (*pglBlendColor)(0.0, 0.0, 0.0, 1.0 - bf);
        glCallList(5);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > .995) || (bf < 0.005) ) {
            bf_inc = -bf_inc;
        } 
    }
}

__declspec(dllexport)
void do_squeeze(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    long x, y;
    float bf = 0;
    float bf_inc;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);
    glDisable(GL_BLEND);
    glLoadIdentity();
    bf_inc = .75;

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glDisable(GL_TEXTURE_2D);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        glEnable(GL_TEXTURE_2D);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer1);
        glRotatef(bf_inc, 1.0, 0.0, 0.0);
        glCallList(1);

        count++;
        (*_SendToVideo) (count);
    }

}


__declspec(dllexport)
void do_perspective(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                    int tex_name, int *playing, int ec)
{
    int val;
    int count = 0;
    long x, y;
    float bf = 0;
    float bf_inc;
    float twist = 0;
    float twist_inc;
    float near_ = .5;
    float trans =  -.95;
    float far_  = 2.0; 
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);
    glDisable(GL_BLEND);

    bf_inc = .20;
    twist_inc = .21;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-0.50, 0.50, -0.50, 0.50, near_, far_);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glClear(GL_COLOR_BUFFER_BIT);
        glDisable(GL_TEXTURE_2D);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        glEnable(GL_TEXTURE_2D);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer1);
        glLoadIdentity();

        glRotatef(bf, 0.0, 1.0, 0.0);
        bf += bf_inc;

        if (count > 100) {
            glRotatef(twist, 1.0, 0.0, 0.0);
            twist += twist_inc;
        }

        glTranslatef(0.0, 0.0, trans);
        glCallList(1);
        count++;
        (*_SendToVideo) (count);
    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

__declspec(dllexport)
void do_depth(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
              int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    long x, y;
    float bf = 0;
    float bf_inc = .125;
    float t1 = 0;
    float t2 = 0;
    float t1_inc = .0035;
    float t2_inc = .0060;
    float near_ = .5;
    float trans =  -0.75;
    float far_  = 2.0; 
    float flip = 1.0;
    char *sourceBuffer1, *sourceBuffer2;

    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex_name);

    glDisable(GL_BLEND);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glClearDepth(1.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-0.70, 0.70, -0.70, 0.70, near_, far_);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    while (*playing != ec) {

        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        glClear(GL_COLOR_BUFFER_BIT);
        glClear(GL_DEPTH_BUFFER_BIT);

        glLoadIdentity();
        glTranslatef(0.0, 0.0, -1.75);
        glTranslatef(0.0, 0.0, t1);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer2);
        glCallList(1);

        glLoadIdentity();
        glTranslatef(0.0, 0.0, trans);
        glTranslatef(0.0, 0.0, t2);

        glRotatef(bf, 1.0, 0.0, 0.0);
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, sourceBuffer1);
        glCallList(1);

        count++;
        (*_SendToVideo) (count);

        i = count % 600;
        if (i == 0) flip = -flip;

        if (i < 275) {
            bf += bf_inc*flip;
        }
        else {
            t2 -= t2_inc*flip;
            t1 += t1_inc*flip;
        }

    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_DEPTH_TEST);

}

__declspec(dllexport)
void do_radial(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
               int tex_name, int *playing, int ec)
{
    /* This routine uses stencil planes to draw a wipe (thanks to the idea from
     * true@boston.sgi.com).  It could be modified easily to draw a large variety
     * of standard SMPTE wipes).
     */
    int i, val;
    long x, y;
    int count = 0;
    float bf = 0.0;
    float bf_inc = .07;
    int step = 1;
    char *sourceBuffer1, *sourceBuffer2;

    glClearStencil(0x0);
    glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    x = tex_width  - uSize;
    y = tex_height - vSize;

    glLoadIdentity();
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);

    while (*playing != ec) {
        (*_GetInputFields) (&sourceBuffer1, &sourceBuffer2);

        /* Draw the background - no stenciling yet */
        glDisable(GL_STENCIL_TEST);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer1);

        /* The selected stencil function will always fail (since
         * the stencil plane is filled with 0's it will never equal
         * 1).  Whenever it fails, have glStencilOp load a 1 into the
         * the stencil buffer.  This loads a 1 into the stencil buffer wherever
         * the rectangle is drawn (and 0 otherwise).
         */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_EQUAL, 0x1, 0x1);
        glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
    
        if (step == 1) {
            glBegin(GL_TRIANGLES);
            glVertex2f(0.0, 0.0);
            glVertex2f(0.0, -1.0);
            glVertex2f(bf, -1.0);
            glEnd();
            bf += bf_inc;
            if (bf > 1.0) {
                step += 1;
                bf = -1.0;
            }
        } else if (step == 2) {
            glBegin(GL_TRIANGLE_FAN);
            glVertex2f(0.0, 0.0);
            glVertex2f(0.0, -1.0);
            glVertex2f(1.0, -1.0);
            glVertex2f(1.0, bf);
            glEnd();
            bf += bf_inc;
            if (bf > 1.0) {
                step += 1;
            }
        } else if (step == 3) {
            glBegin(GL_TRIANGLE_FAN);
            glVertex2f(0.0, 0.0);
            glVertex2f(0.0, -1.0);
            glVertex2f(1.0, -1.0);
            glVertex2f(1.0, 1.0);
            glVertex2f(bf, 1.0);
            glEnd();
            bf -= bf_inc;
            if (bf < -1.0) {
                step += 1;
                bf = 1.0;
            }
        } else if (step == 4) {
            glBegin(GL_TRIANGLE_FAN);
            glVertex2f(0.0, 0.0);
            glVertex2f(0.0, -1.0);
            glVertex2f(1.0, -1.0);
            glVertex2f(1.0, 1.0);
                glVertex2f(-1.0, 1.0);
            glVertex2f(-1.0, bf);
            glEnd();
            bf -= bf_inc;
            if (bf < -1.0) {
                step += 1;
            }
        } else if (step == 5) {
            glBegin(GL_TRIANGLE_FAN);
            glVertex2f(0.0, 0.0);
            glVertex2f(0.0, -1.0);
            glVertex2f(1.0, -1.0);
            glVertex2f(1.0, 1.0);
            glVertex2f(-1.0, 1.0);
            glVertex2f(-1.0, -1.0);
            glVertex2f(bf, -1.0);
            glEnd();
            bf += bf_inc;
            if (bf > 0.0) {
                step = 1;
                bf = 0.0;
            }
        } 

        /* The stencil function is still "equals 1", so wherever the rectangle
         * has been drawn, the glDrawPixels will show through.  When the
         * stencil test passes reset the values in the stencil buffer to
         * 0 effectively clearing the stencil buffer for the next time.
         */ 
        glStencilOp(GL_KEEP, GL_KEEP, GL_ZERO);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, sourceBuffer2);

        count++;
        (*_SendToVideo) (count);

    }
    glDisable(GL_STENCIL_TEST);
}

__declspec(dllexport)
void do_3stream_tex(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                    int tex_name, int *playing, int ec)
{
    int i, val;
    long x, y;
    int count = 0;
    float bf = 0;
    float bc = 0;
    float bf_inc;
    float bc_inc;
    char *fields[3];

    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT);
    x = tex_width  - uSize;
    y = tex_height - vSize;

    glBlendFunc(GL_CONSTANT_ALPHA_EXT,GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
    (*pglBlendEquation)(GL_FUNC_ADD_EXT);

    bf_inc = .01;
    bc_inc = .01;

    while (*playing != ec) {

        (*_GetInputArray) (3, fields);

        glDisable(GL_TEXTURE_2D);
        glDisable(GL_BLEND);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[0]);

        glEnable(GL_TEXTURE_2D);
        glEnable(GL_BLEND);

        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, fields[1]);
        glLoadIdentity();
        glScalef(0.5, 0.5, 0.5);
        glTranslatef(bf, 0.0, 0.0);
        (*pglBlendColor)(0.0, 0.0, 0.0, bc);
        glCallList(1);

        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, fields[2]);
        glLoadIdentity();
        glScalef(0.5, 0.5, 0.5);
        glTranslatef(-bf, 0.0, 0.0);
        (*pglBlendColor)(0.0, 0.0, 0.0, 1.0 - bc);
        glCallList(1);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 1.995) || (bf < -1.995) ) {
            bf_inc = -bf_inc;
        }

        bc += bc_inc;
        if ((bc > .995) || (bc < 0.005) ) {
            bc_inc = -bc_inc;
        }

    }
}

__declspec(dllexport)
void do_3stream_wipe(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                     int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    float bf;
    float bc;
    float bf_inc;
    float bc_inc;
    char *fields[3];

    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT);

    glBlendFunc(GL_CONSTANT_ALPHA_EXT,GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
    (*pglBlendEquation)(GL_FUNC_ADD_EXT);
    glDisable(GL_TEXTURE_2D);

    bf = -0.9;
    bf_inc = .01;

    bc = 0.0;
    bc_inc = .01;

    while (*playing != ec) {

        (*_GetInputArray) (3, fields);

        /* Draw the background - no stenciling yet */
        glDisable(GL_BLEND);
        glDisable(GL_STENCIL_TEST);

        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[0]);

        /* Clear the stencil plane */
        glClearStencil(0x0);
        glClear(GL_STENCIL_BUFFER_BIT);

        /* The selected stencil function will always fail (since
         * the stencil plane is filled with 0's it will never equal
         * 1).  Whenever it fails, have glStencilOp load a 1 into the
         * the stencil buffer.  This loads a 1 into the stencil buffer wherever
         * the rectangle is drawn (and 0 otherwise).
         */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_EQUAL, 0x1, 0x1);
        glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
        glRectf(-1.0, -1.0, bf, 1.0);

        /* The stencil function is still "equals 1", so wherever the rectangle
         * has been drawn, the glDrawPixels will show through.
         */
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[1]);


        /* Now draw third stream on top with blending */
        glDisable(GL_STENCIL_TEST);
        glEnable(GL_BLEND);
        (*pglBlendColor)(0.0, 0.0, 0.0, bc);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[2]);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 0.9) || (bf < -0.9) ) {
            bf_inc = -bf_inc;
        }

        bc += bc_inc;
        if ((bc > .995) || (bc < 0.005) ) {
            bc_inc = -bc_inc;
        }

    }
    glDisable(GL_STENCIL_TEST);
}

__declspec(dllexport)
void do_4stream(int uSize, int vSize, int tex_width, int tex_height, int tex_format, 
                int tex_name, int *playing, int ec)
{
    int i, val;
    int count = 0;
    long x, y;
    float bf;
    float bc;
    float bf_inc;
    float bc_inc;
    char *fields[4];

    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT);

    x = tex_width  - uSize;
    y = tex_height - vSize;

    glBlendFunc(GL_CONSTANT_ALPHA_EXT,GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
    (*pglBlendEquation)(GL_FUNC_ADD_EXT);

    bf = -0.9;
    bf_inc = .01;

    bc = 0.0;
    bc_inc = .01;

    while (*playing != ec) {

        (*_GetInputArray) (4, fields);

        /* Draw the background - no stenciling yet */
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_BLEND);
        glDisable(GL_STENCIL_TEST);
        glLoadIdentity();

        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[0]);

        /* Clear the stencil plane */
        glClearStencil(0x0);
        glClear(GL_STENCIL_BUFFER_BIT);

        /* The selected stencil function will always fail (since
         * the stencil plane is filled with 0's it will never equal
         * 1).  Whenever it fails, have glStencilOp load a 1 into the
         * the stencil buffer.  This loads a 1 into the stencil buffer wherever
         * the rectangle is drawn (and 0 otherwise).
         */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_EQUAL, 0x1, 0x1);
        glStencilOp(GL_REPLACE, GL_KEEP, GL_KEEP);
        glRectf(-1.0, -1.0, bf, 1.0);

        /* The stencil function is still "equals 1", so wherever the rectangle
         * has been drawn, the glDrawPixels will show through.
         */
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
        glDrawPixels(uSize, vSize, tex_format, GL_UNSIGNED_BYTE, fields[1]);

        /* Now draw third and fourth streams on top with blending */
        glDisable(GL_STENCIL_TEST);
        glEnable(GL_BLEND);
        glEnable(GL_TEXTURE_2D);

        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, fields[2]);
        glLoadIdentity();
        glScalef(0.5, 0.5, 0.5);
        glTranslatef(bf, 0.0, 0.0);
        (*pglBlendColor)(0.0, 0.0, 0.0, bc);
        glCallList(1);

        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, uSize, vSize, tex_format, 
                           GL_UNSIGNED_BYTE, fields[3]);
        glLoadIdentity();
        glScalef(0.5, 0.5, 0.5);
        glTranslatef(-bf, 0.0, 0.0);
        (*pglBlendColor)(0.0, 0.0, 0.0, 1.0 - bc);
        glCallList(1);

        count++;
        (*_SendToVideo) (count);

        bf += bf_inc;
        if ((bf > 0.9) || (bf < -0.9) ) {
            bf_inc = -bf_inc;
        }

        bc += bc_inc;
        if ((bc > .995) || (bc < 0.005) ) {
            bc_inc = -bc_inc;
        }

    }
    glDisable(GL_STENCIL_TEST);
}
