/*
 * Copyright (c) 1998 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* ntfio.c = Tests read/write performance with unbuffered IO. 
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-n ntries>         - Number of read or write operations to perform
 *			     (default = 250). 
 *     <-size megaBytes>   - Number of megabytes per operation (default = 4). 
 *     <-w>	           - Perform write operations (default is read).
 *     fileName            - file to read or write. 
 *
 */


#include <stdio.h>
#include <time.h>
#include <windows.h>
#include <iostream.h>


#define READ_OP  1
#define WRITE_OP 2

main(int argc, char *argv[])
{
    DWORD ntries = 250;	      /* Default to 250 operations */
    DWORD bytes_per_try = 4;  /* Try 4 meg operations.     */
    DWORD op = READ_OP;	      /* Default to reads.	   */ 
    DWORD i, j;
    DWORD sectorsPerCluster, bytesPerSector, freeClusters, nClusters;
    DWORD bytesMoved;
    HANDLE fd;
    LPCTSTR Fname;
    char *data;
    double tdiff;
    double bytes;
    clock_t start, finish;

    /* Parse command line input */
    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            ntries = atoi(argv[i]);
        }
        else if (strcmp (argv[i], "-size") == 0) {
            ++i;
            bytes_per_try = atoi(argv[i]);
        }
        else if (strcmp (argv[i], "-w") == 0) {
	    op = WRITE_OP;
        }
        else  {
            Fname = argv[i];
        }
    }

    bytes_per_try = 1024 * 1024 * bytes_per_try;

#if 0
    /* bytes_per_try must be a multiple of the bytes per sector of the volume
     * to be used.  Unfortunately, GetDiskFreeSpace() (which returns the bytesPerSector
     * value) often fails for no known reason.  bytesPerSector is usually 512. 
     */
    bytesPerSector = 0;
    GetDiskFreeSpace("F:", &sectorsPerCluster, &bytesPerSector, &freeClusters, &nClusters);
    fprintf(stderr, "Bytes per sector = %d\n", bytesPerSector);

    if ((bytes_per_try % bytesPerSector) != 0) {
        fprintf(stderr, "%d not multiple of %d\n", bytes_per_try, bytesPerSector);
        exit (-1);
    }
#endif


    /* Allocate memory for the read/write operation */
    if ((data = (char *)VirtualAlloc(NULL, bytes_per_try, MEM_COMMIT, PAGE_READWRITE)) == NULL) {
        fprintf(stderr, "VirtualAlloc failed\n");
        exit (-1);
    }

    memset (data, 42, bytes_per_try);

    /* Create or open the file for unbuffered IO */
    if (op == READ_OP) {
        fprintf(stderr, "***** READING *****\n");
        fd = CreateFile(Fname, GENERIC_READ, 0, NULL, OPEN_EXISTING, 
                        FILE_FLAG_NO_BUFFERING, NULL);
    }
    else {
        fprintf(stderr, "***** WRITING *****\n");
        fd = CreateFile(Fname, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 
                        FILE_FLAG_NO_BUFFERING, NULL);
    }

    if (fd  == INVALID_HANDLE_VALUE) {
    	fprintf(stderr, "Failed to open %s\n", Fname);
        exit (-1);
    }

    fprintf(stderr, "Doing direct IO to move %d bytes %d times to %s\n",
                     bytes_per_try, ntries, Fname);


    /* Loop the requested number of times, performin a read or a write
     * operation.
     */
    start = clock();

    for (i=0; i < ntries; i++) {

        if (op == READ_OP) {
            j = ReadFile(fd, data, bytes_per_try, &bytesMoved, NULL);
        }
        else {
            j = WriteFile(fd, data, bytes_per_try, &bytesMoved, NULL);
        } 

        /* Check for errors or going past end of file */
        if ((j == 0) || (bytesMoved != bytes_per_try)) {
	    fprintf(stderr, "Operation %d failed,  status = %d moved = %d\n",
	                     i, j, bytesMoved);
	    exit (-1);
        }

        if (i%25 == 0) fprintf(stderr, "Done %d operations\n", i);
    }

    finish = clock();

    tdiff = (double)(finish - start)/CLOCKS_PER_SEC;
    bytes = (double)bytes_per_try * (double)ntries;
    fprintf(stderr, "Total time  = %f\n", tdiff);
    fprintf(stderr, "Throughput  = %f MB/sec\n", (bytes/tdiff)/1000000.0);

    CloseHandle(fd);
}
