/*
 * Copyright (c) 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* pdv_streamfx_u -Plays up to four uncompressed streams from disk, performs an 
 *			       effect in an off screen pbuffer that is read back into memory 
 *			       and sent to video-out.
 *
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-single>		- use single buffering 
 *     <-iov>		    - number of samples read per ReadFile() call. 
 *     <-gfx_size x y>  - Size of gfx window on scree, defaults to sample size. 
 *     <-gfx_rate #>    - Update rate of gfx (by default it is not updated).
 *                        Value of n --> update gfx monitor with nth sample. 
 *     file1			- file for first stream (created with vgrab_u). 
 *     <file2>			- file for 2nd stream   (created with vgrab_u). 
 *     <file3>			- file for 3rd stream   (created with vgrab_u). 
 *     <file4>			- file for 4th stream   (created with vgrab_u). 
 *
 * The input files may contain GL_RGB (24 bits per pixel), GL_RGBA (32 bits 
 * per pixel), or GL_YCRCB_SGIX (yuv 4:2:2 data @ 16 bits per pixel).
 *
 *
 * How to run the program:
 * =======================
 * When you run the program, a blank window appears.  Click once with the left 
 * mouse to start play, click again with the same button to stop play.  The 
 * initial effect performed is a blend.  Click at any time with the right mouse
 * button to change the effect being played.  See the routine "loadit" for 
 * effect choices. After entering the new effect, play resumes immediately.  
 * The effect calculation routines are inside common/fxcode.c 
 *
 * Overview of program flow: 
 * =========================
 * 
 * Each time you click the button and start play, a reader thread is started
 * for each file (see routine ReadIt) and an effect processing thread (see
 * routine PlayEffect) is also started.  The effect thread retrieves samples
 * for each stream from output queues associated with each reader thread 
 * (see stream_t).  The effect thread performs the desired effect in an off-screen
 * scree pbuffer.  The effect is automatically read back into system memory 
 * (see _videBufs array in SendToVideo) and sent to video out. 
 *
 * All threads are started from within the routine doit(), which is entered
 * whenever a button is depresed.  doit() also stops play as well.
 *
 * Here are the thread descriptions:
 * =================================
 *    1. UI thread - This threasd sits in a loop waiting for a
 *		     button press.  If a left button press, then the UI enters
 *		     the routine doit().  Inside doit() play is either stopped
 *		     or started depending upon the current play state.  If the
 *		     right mouse button is pressed, then loadit() is entered.
 *		     The user is prompted for a new effect, and play is resumed.
 *		     The UI thread directs the effect thread to exit whenever 
 *		     play is stopped. 
 *
 *    2. Readit() - One of these threads is created per file.  The reader
 *		    threads read uncompressed data from disk and place the
 *		    the data in the "readvBufs" buffer associated with each
 *		    stream (see stream_t data structure).  The number of fields
 *		    that can be held in "readvBufs" depends on the size of the 
 *		    io vector being used with readv.  Each "readvBufs" is 
 *		    implemented as a circular buffer, with reader threads adding 
 *		    fields at "rvtail" and the effect thread removing fields from
 *		    "rvhead" (these are both members of stream_t). 
 *
 *		    The reader threads use unbuffered I/O when reading from disk.
 *		    Each field in the input file has been padded out to a 16k
 *		    boundary when captured (see vgrab_u). The optional parameter
 *		    -iov specifies the numbers of samples read per ReadFile() call.
 *		    In order to simplify the read code, the number of fields in each
 *		    file is truncated to a multiple of the io vector.
 *
 *    3. PlayEffect() - This is the thread which computes the effect. The effect is 
 *		    calculated in an off-screen pbuffer, usually one stream is drawn 
 *		    as backgroud via glDrawPixles and the other is texture mapeed.  
 *		    The effect "depth" texture maps both streams.  Look in ../common/fxcode.c
 *          for the effetc routines.  The resulting effect is read back into memory 
 *          and sent to video out (see SendToVideo). 
 *
 *		    This thread waits for valid output from each of the reader threads 
 *          by examining the rvhead variable in the desired stream_t structure.  
 *          After it has a  valid field for each stream it goes ahead and calculates 
 *          the effect. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <windows.h>
#include <time.h>
#include <GL\gl.h>
#include <sgidmedia.h>
#include "..\common\handy.h"
#include "..\common\dmhandy.h"
#include "..\common\fileutils.h"
#include "..\common\filepacking.h"
#include "..\common\fxcode.h"

typedef GLuint (APIENTRY * PFNWGLCREATEPBUFFEREXTPROC)(HDC, GLuint, GLsizei, GLsizei, int *); 
typedef BOOL   (WINAPI   * PFNWGLMAKECURRENTREADSGI)  (HDC, HDC, HGLRC);
PFNWGLCREATEPBUFFEREXTPROC pwglCreatePBufferSGIX;
PFNWGLMAKECURRENTREADSGI   pwglMakeCurrentReadSGI;

// When we send buffers to dmPath via dmPathSendBuffers, use
// the following indices for the image buffer, UST, and MSC.
#define _VC_BUFFER_INDEX    0
#define _VC_UST_INDEX       1
#define _VC_MSC_INDEX       2
#define _VC_BUFFER_PV_END   3
#define _RVBUFS             8  //How many buffers for driving video
#define MAX_IOV             24 

/* One stream structure per file */
typedef struct stream {
    char *readvBufs;    /* Each ReadIt() thread deposits fields in this buffer */
    int   rvhead;       /* Where PlayEffect() retrieves fields from.           */
    int   rvtail;       /* Where Readit() places fields it has read from disk. */
    int   num_rvbufs;   /* Number of fields in readvBufs                       */
    int   curField;     /* The current field being processed from the file.    */
    int iovcnt;         /* The size of io vector for this stream.              */
    fileHeader_t f;
} stream_t;


/* wgl defines */
HDC             _deviceContext;    //device for on screen window.
HGLRC           _renderContext;    //render context for on screen window.
HDC             _PBdeviceContext;  //device for PBUFFER. 
HGLRC           _PBrenderContext;  //render context for PBUFFER. 

/* File defines */
int _nreaders;
int _nstreams = 0;
int _uSize;
int _vSize;
stream_t *_s;

/* texture defines */
unsigned int _texA;                      /* texture names */
int _tex_width;                          /* texture size */
int _tex_height;                         /* texture size */
int _tex_format;
FxType_t _fx = FX_WIPE;

/* GFX defines */
int _dbuffer = 1;
int _gfx_rate  = 0;
int _winx = 0;
int _winy = 0;

/* Play control variables */
int _IamPlaying = 0;
int _count;

/* video buffer declarations  */
char    **_videoBufs;	/* Array of buffer pointers for sending results to video. */
int       _videoSize;
char      _bufStatus[_RVBUFS];
int       _mscStep;
HDMPATH   _hVideoPath;

/* External subroutines: look in ..\common\common.dll */
void SetupConsole();
void createTexture(unsigned int *texname, int tex_width, int tex_height, int tex_format, int mip);
void SetVideoSendRoutine(void (*f)());
void SetVideoGetFieldsRoutine(void (*f)());
void SetVideoGetArrayRoutine(void (*f)());
void LoadGLExtensions();
int  FilePackingToDMFormat(int packing);
int  BuildVideoLists(int start, int u, int v, int tw, int th);

/* Local subroutines    */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void Quit(HWND);
void initGL(HWND);
void doit();
void exposeCB(HWND, PAINTSTRUCT *); 
void loadit();
DWORD WINAPI PlayEffect(LPVOID);
DWORD WINAPI ReadIt(LPVOID);
void GetInputFields(char **f1, char **f2);
void GetInputArray (int nfields, char *fields[]);
void SendToVideo (int count);
void initVideo(dmJackTypeEnum);
int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc);
void initStreamRead(int iovcnt, stream_t *s);
int GetNumReaders(FxType_t fx);

/* Main entry point from windows */

int APIENTRY WinMain(HINSTANCE _instance, HINSTANCE _prevInst, LPSTR _cmdLine,
                     int _cmdShow)
{
    int i,n;
    int iovcnt = 4;
    char *f[6];
    char *bigBuffer;
    MSG msg ;
    WNDCLASSEX wndClass;
    HWND wnd;
    char *className = "OpenGL";
    char *windowName = "pdv_streamfx_u";

    /* Set up standard IO for printing to a console window.*/
    SetupConsole();

    for (i=1; i < __argc; i++) {
        if (strcmp (__argv[i], "-single") == 0) {
            _dbuffer = 0; 
        }
        else if (strcmp (__argv[i], "-iov") == 0) {
            ++i;
            iovcnt = atoi(__argv[i]);
            if (iovcnt > MAX_IOV) iovcnt = MAX_IOV;
        }
        else if (strcmp (__argv[i], "-gfx_rate") == 0) {
            ++i;
            _gfx_rate = atoi(__argv[i]);
        }
        else if (strcmp (__argv[i], "-gfx_size") == 0) {
            ++i;
            _winx = atoi(__argv[i]);
            ++i;
            _winy = atoi(__argv[i]);
        }
        else {
            f[_nstreams]  = __argv[i];
            _nstreams++;
        }
    }

    if (_nstreams == 1) {
        _fx = FX_NONE;
    }

    NC (_s = (stream_t *)VirtualAlloc(NULL, _nstreams*sizeof(stream_t), MEM_COMMIT, PAGE_READWRITE));

    /* Open up file streams */
    for (i=0; i < _nstreams; i++) {

        fileInit(f[i], &(_s[i].f), TRUE);
        initStreamRead(iovcnt, &(_s[i]));

        /* Make sure we have an acceptable packing.     */
        if ((_s[i].f.packing != VL_PACKING_YVYU_422_8)  &&
            (_s[i].f.packing != VL_PACKING_BGR_8_P) &&
            (_s[i].f.packing != VL_PACKING_ABGR_8) ) {
            ERR(1,"Bad packing");
        }

        ERR ((_s[i].f.packing != _s[0].f.packing), "Mismatch in file packing");
        ERR ((_s[i].f.uSize   != _s[0].f.uSize),   "Mismatch in file widths");
        ERR ((_s[i].f.vSize   != _s[0].f.vSize),   "Mismatch in file heights");
    }

    _uSize = _s[0].f.uSize;
    _vSize = _s[0].f.vSize;
    fprintf(stderr, "Playing %dx%d read size = %d iov = %d\n",
                     _uSize, _vSize, _s[0].f.readSize, _s[0].iovcnt);


    //Allocate buffers for sending to video
	_videoSize = _uSize * _vSize * 4;

	_videoBufs =  (char **) VirtualAlloc(NULL,  _RVBUFS*sizeof(char *), MEM_COMMIT,PAGE_READWRITE);
	bigBuffer  =  (char * ) VirtualAlloc(NULL, _RVBUFS * _videoSize, MEM_COMMIT,PAGE_READWRITE);
	ERR ((bigBuffer == NULL), "Failed to allocate reader buffers");

    for (i=0; i < _RVBUFS; i++) {
	    _videoBufs[i] = bigBuffer + i*_videoSize;
    }

	// Init video hardware.
	initVideo(DM_JACK_TYPE_COMPOSITE);

    // Check window dimensions.
    if (_winx == 0) {
        _winx = _s[0].f.uSize;
        _winy = _s[0].f.vSize;
    }

    // Do that nasty windowing stuff.
    wndClass.cbSize        = sizeof (WNDCLASSEX) ;
    wndClass.style         = CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc   = WndProc;
    wndClass.cbClsExtra    = 0 ;
    wndClass.cbWndExtra    = 0 ;
    wndClass.hInstance     = _instance ;
    wndClass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
    wndClass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
    wndClass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wndClass.lpszMenuName  = NULL ;
    wndClass.lpszClassName = className ;
    wndClass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

    /* Create window */
    RegisterClassEx (&wndClass) ;
    wnd = CreateWindow( className, windowName,
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                            100,                            // initial x position
                            100,                            // initial y position
                            _winx,                          // width
                            _winy + 22,                     // Height
                            NULL,                           // parent window handle
                            (HMENU) NULL,                   // window menu handle
                            _instance,                      // program instance handle
                            NULL) ;

    ShowWindow(wnd, _cmdShow);
    UpdateWindow(wnd);

    /* Now loop here and process messages to the window.  */
    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg) ;
        DispatchMessage (&msg) ;
    }

    return msg.wParam ;
}




LRESULT CALLBACK WndProc(HWND _wnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
    /* Routine to process window events.	*/
    PAINTSTRUCT ps;

    // Process the event.
    switch(_msg) {
        case WM_CREATE:
            initGL(_wnd);
            return 0;

        case WM_LBUTTONUP:
        case WM_MBUTTONUP:
            doit();
            return 0;

        case WM_RBUTTONUP:
            loadit();
            return 0;

        case WM_SYSCOMMAND:
            if(_wParam == SC_CLOSE) { 
                fprintf(stderr, "Quitting ...\n");
                Quit(_wnd); 
                return 0; 
            }
            else break;

        case WM_PAINT:
            exposeCB(_wnd, &ps);
            return 0;
    }

    return DefWindowProc (_wnd, _msg, _wParam, _lParam) ;
}


void initGL(HWND wnd)
{
    int i, val;
    float xscale, yscale;
    int     pixelFormat;
    PIXELFORMATDESCRIPTOR   pfd;

    memset (&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
    pfd.nSize =                 sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion =              1;
    pfd.dwFlags =               PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW;
    pfd.iPixelType =            PFD_TYPE_RGBA;
    pfd.cColorBits =            24;
    pfd.cDepthBits =            24;
    pfd.cStencilBits =          1;
    pfd.iLayerType =            PFD_MAIN_PLANE;

    // Now set pfd for on screen window.
    if (_dbuffer) {
        pfd.dwFlags |= PFD_DOUBLEBUFFER;
    }

    // Get the device context for the window.
    NC (_deviceContext = GetDC(wnd));

    // Match a supported pixel format to the specified pixel format.
    pixelFormat = ChoosePixelFormat(_deviceContext, &pfd);

    // Set the pixel format of the device context.
    ZC (SetPixelFormat(_deviceContext, pixelFormat, &pfd)); 

    // Create OpenGL rendering context and make it current.
    NC (_renderContext   = wglCreateContext(_deviceContext));
    NC (_PBrenderContext = wglCreateContext(_deviceContext));
    ZC (wglMakeCurrent(_deviceContext, _renderContext));

    // init on screen window. 
    glViewport(0, 0, _winx, _winy);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glClearColor(0.0,0.0,0.0,1.0);
    glColor4f(1.0, 1.0, 1.0, 1.0);

    glRasterPos2f(-1.0, 1.0);  
    xscale = ((float)_winx)/_s[0].f.uSize;
    yscale = ((float)_winy)/_s[0].f.vSize;
    glPixelZoom(xscale, -yscale);


    ZC (wglMakeCurrent(_deviceContext, _PBrenderContext));
    pwglCreatePBufferSGIX =  (PFNWGLCREATEPBUFFEREXTPROC)wglGetProcAddress("wglCreatePBufferSGIX");
    pwglMakeCurrentReadSGI=  (PFNWGLMAKECURRENTREADSGI  )wglGetProcAddress("wglMakeCurrentReadSGI");
    //fprintf(stderr, "Function pointers = %x %x\n",pwglCreatePBufferSGIX, pwglMakeCurrentReadSGI);


    // Create Pbuffer
    memset (&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
    pfd.nSize =                 sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion =              1;
    pfd.dwFlags =               PFD_SUPPORT_OPENGL;
    pfd.iPixelType =            PFD_TYPE_RGBA;
    pfd.cColorBits =            24;
    pfd.cDepthBits =            24;
    pfd.cStencilBits =          1;
    pfd.iLayerType =            PFD_MAIN_PLANE;
    pixelFormat = ChoosePixelFormat(_deviceContext, &pfd);


    _PBdeviceContext = (HDC)(*pwglCreatePBufferSGIX)(_deviceContext, (GLuint)pixelFormat, 
                                                     _uSize, _vSize, NULL);
    ERR((_PBdeviceContext == 0), "Can't create PBUFFER"); 


    // Make Pbuffer device and context current.
    ZC ((*pwglMakeCurrentReadSGI)(_PBdeviceContext, _PBdeviceContext, _PBrenderContext));

    /* Set up texture type, width, and size */
    if (_s[0].f.packing == VL_PACKING_ABGR_8)         _tex_format = GL_RGBA;
    else if (_s[0].f.packing == VL_PACKING_BGR_8_P)   _tex_format = GL_RGB;
    else if (_s[0].f.packing == VL_PACKING_YVYU_422_8)_tex_format = GL_YCRCB_SGIX;

    _tex_width = 1024; 
    _tex_height = (_vSize <= 256) ? 256 : 512; 

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
 
    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_format, 0);

    /* init viewport */
    glViewport(0, 0, _s[0].f.uSize, _s[0].f.vSize);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    glDisable(GL_DITHER);

    (void)BuildVideoLists(1, _uSize, _vSize, _tex_width, _tex_height);
    glRasterPos2f(-1.0, -1.0);  /* Draw upside down for now */

    if (_tex_format ==  GL_YCRCB_SGIX) {
        fprintf(stderr, "***** Setting OpenGL Pixel Unpacking for YUV 422 *******\n");
        glPixelStorei(GL_UNPACK_SUBSAMPLE_RATE_SGIX, GL_PIXEL_SUBSAMPLE_4242_SGIX);
    }
    
    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR INIT ******* 0x%x\n",val);
    }

    LoadGLExtensions();


}




void exposeCB(HWND wnd, PAINTSTRUCT *ps) 
{
    BeginPaint(wnd, ps);
    if (_IamPlaying == 0) {

        ZC (wglMakeCurrent(_deviceContext, _renderContext));

        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) SwapBuffers(_deviceContext);

        glClear(GL_COLOR_BUFFER_BIT);
        glFlush();
        fprintf(stderr, "exposeCB\n");
    }
    EndPaint(wnd, ps);
}




void doit()
{
    int allDone,i;
    HANDLE playThread, readT;
    DWORD  playThreadID, readID;
    double duration;
    static clock_t start, finish;

    /* Routine to START/STOP play. 
     * This routine is entered directly when the left mouse button is clicked
     * or from loadit() when changing effects.
     */

    if (_IamPlaying) {
        /* Then stop playing.	*/
        _IamPlaying = 0;
        allDone = -(_nreaders + 1);     // need to wait for readers + the player.

        finish = clock();
        
        /* Wait for both readers and the effect thread to exit. */
        while (_IamPlaying != allDone) Sleep (10);
        _IamPlaying = 0;
 
        duration = (double)(finish - start)/CLOCKS_PER_SEC;
        fprintf(stderr, "Draw time = %f sec. fps = %f\n", duration, ((double)(_count))/duration);

        ZC (wglMakeCurrent(_deviceContext, _renderContext));
        ZC ((*pwglMakeCurrentReadSGI)(_deviceContext, _PBdeviceContext, _renderContext));

        glCopyPixels(0, 0, _s[0].f.uSize, _s[0].f.vSize, GL_COLOR);
        if (_dbuffer) SwapBuffers(_deviceContext);

        glCopyPixels(0, 0, _s[0].f.uSize, _s[0].f.vSize, GL_COLOR);
        glFlush();
    }
    else {
        fprintf(stderr, "Starting play\n");
        _IamPlaying = 1;
        _nreaders = GetNumReaders(_fx);

        // Surrender context asociated with PBuffer.
        ZC (wglMakeCurrent(NULL, NULL));
        ZC ((*pwglMakeCurrentReadSGI)(NULL, NULL, NULL));

        // Set status for video read back buffers.
        for (i=0; i < _RVBUFS; i++) _bufStatus[i] = 0;

        //Start all reader threads.
        for (i=0; i < _nreaders; i++) {
            _s[i].rvhead = _s[i].rvtail = 0;
            NC(readT = CreateThread (NULL, 0, ReadIt, (LPVOID *)&(_s[i]), 0, &readID));
        }

        // Start the effect play back thread.
        NC(playThread = CreateThread (NULL, 0, PlayEffect, NULL, 0, &playThreadID));
        start = clock();
    }

}



void loadit()
{
    /* This routiine is entered whenever a new effect is loaded.  This
     * is done by pushing the right mouse button.
     */ 
    char etype[16];

    if (_fx == FX_NONE) return;

    /* Stop play */
    if (_IamPlaying) doit();

    fprintf(stderr, "Enter effect: ");
    gets(etype);

    if (strcmp (etype,      "blend") == 0)    _fx = FX_BLEND;
    else if (strcmp (etype, "pip") == 0)      _fx = FX_PIP;
    else if (strcmp (etype, "wipe") == 0)     _fx = FX_WIPE;
    else if (strcmp (etype, "cross") == 0)    _fx = FX_CROSS;
    else if (strcmp (etype, "boxblend") == 0) _fx = FX_BOXBLEND;
    else if (strcmp (etype, "squeeze") == 0)  _fx = FX_SQUEEZE;
    else if (strcmp (etype, "persp") == 0)    _fx = FX_PERSPECTIVE;
    else if (strcmp (etype, "depth") == 0)    _fx = FX_DEPTH;
    else if (strcmp (etype, "radial") == 0)   _fx = FX_RADIAL;
    else if (strcmp (etype, "3stream_wipe") == 0)  _fx = (_nstreams >= 3) ? FX_3STREAM_WIPE : _fx;
    else if (strcmp (etype, "3stream_tex") == 0)   _fx = (_nstreams >= 3) ? FX_3STREAM_TEX  : _fx;
    else if (strcmp (etype, "4stream") == 0)       _fx = (_nstreams == 4) ? FX_4STREAM      : _fx;

    /* Resume play */
    doit();
}


DWORD WINAPI PlayEffect(LPVOID params)
{
    /* This is the effect calculation thread launched from doit()
     * It jumps off (for the entire current play duration) into
     * one of the effect routines.
     */

    int ec = -_nreaders; 
    int i;
    int count;
    char *dataPtr;
    DMint64 ust, msc;

    ZC(wglMakeCurrent(_deviceContext, _PBrenderContext));
    ZC((*pwglMakeCurrentReadSGI)(_PBdeviceContext, _PBdeviceContext, _PBrenderContext));

    // These function pointers are required by do_* routines found
    // in ..\common\fxcode.c
    //
    SetVideoSendRoutine(SendToVideo);
    SetVideoGetFieldsRoutine(GetInputFields);
    SetVideoGetArrayRoutine(GetInputArray);


    switch (_fx) {
	case FX_NONE:        
        do_single(_uSize, _vSize, _tex_width, _tex_height,
                  _tex_format, _texA, &_IamPlaying,ec);
        break;
	case FX_BLEND:       
        do_blend(_uSize, _vSize, _tex_width, _tex_height, 
             _tex_format, _texA, &_IamPlaying,ec);       
        break;
	case FX_WIPE :       
        do_wipe(_uSize, _vSize, _tex_width, _tex_height, 
                _tex_format, _texA, &_IamPlaying,ec);        
        break;
	case FX_PIP  :       
        do_pip(_uSize, _vSize, _tex_width, _tex_height, 
               _tex_format, _texA, &_IamPlaying,ec);         
        break;
	case FX_CROSS:       
        do_cross(_uSize, _vSize, _tex_width, _tex_height, 
                 _tex_format, _texA, &_IamPlaying,ec);       
        break;
	case FX_BOXBLEND:    
        do_boxblend(_uSize, _vSize, _tex_width, _tex_height, 
                    _tex_format, _texA, &_IamPlaying,ec);    
        break;
	case FX_SQUEEZE:     
        do_squeeze(_uSize, _vSize, _tex_width, _tex_height, 
                   _tex_format, _texA, &_IamPlaying,ec);     
        break;
	case FX_PERSPECTIVE: 
        do_perspective(_uSize, _vSize, _tex_width, _tex_height, 
                       _tex_format, _texA, &_IamPlaying,ec); 
        break;
	case FX_DEPTH:       
        do_depth(_uSize, _vSize, _tex_width, _tex_height, 
                 _tex_format, _texA, &_IamPlaying,ec);       
        break;
	case FX_RADIAL:       
        do_radial(_uSize, _vSize, _tex_width, _tex_height, 
                  _tex_format, _texA, &_IamPlaying,ec);       
        break;
    case FX_3STREAM_WIPE:
        do_3stream_wipe(_uSize, _vSize, _tex_width, _tex_height,
                        _tex_format, _texA, &_IamPlaying,ec);
        break;
    case FX_3STREAM_TEX:
        do_3stream_tex(_uSize, _vSize, _tex_width, _tex_height,
                        _tex_format, _texA, &_IamPlaying,ec);
        break;
    case FX_4STREAM:
        do_4stream(_uSize, _vSize, _tex_width, _tex_height,
                        _tex_format, _texA, &_IamPlaying,ec);
        break;
    }

    /* Surrender the graphics context.  */
    glFlush();
    ZC (wglMakeCurrent(NULL, NULL));
    ZC ((*pwglMakeCurrentReadSGI)(NULL, NULL, NULL));
   
    /* Shut down video */ 
    fprintf(stderr, "Ending video transfer\n");
    DC (dmPathEndTransfer(_hVideoPath) );
    DC (dmPathGetReceiveQueueMessageCount(_hVideoPath, &count)); 
        
    for (i = 0; i < count; i++) {
        checkReceiveQueue(&dataPtr, &ust, &msc);
    }

    --_IamPlaying;
    ExitThread (-1); 
}

void SendToVideo (int count)
{

    /* This is the routine called by the effect code to send buffers to
     * video out.  _videBufs[1.._RVBUFS] is the buffer pool driving video.
     * Note:
     *    _bufStatus[i] == 1 --> _videoBufs[i] is still in use by driver. 
     *    _videoBufs[playBuf] is the next buffer to be sent to video. 
     *    _videoBufs[curBuf]  is the next buffer due back from video. 
     *    
     */

    static DMint64 oldmsc;
    static DMint64 newmsc;
    static int playBuf, curBuf;
    DMint64 ust;
    DMpv p[10];
    char *buf, *dataPtr;
    int i, tmphead;

    for (i = 0; i < _nreaders; i++) {
        /* Tell reader we are done with the current set of fields. */
        tmphead = _s[i].rvhead + 1;
        if (tmphead == _s[i].num_rvbufs) tmphead = 0;
        _s[i].rvhead = tmphead;
    }


    _count = count;

    // Check if first time in routine since a new play started.
    if (count == 1) {
        oldmsc = -1;
        curBuf = 0;
        playBuf= 0;
    }


    do {
        // Check and see if previous buffers are available for reuse.
        // Need to check each passs to avoid deadlock with ReadIt().
        checkReceiveQueue(&dataPtr, &ust, &newmsc);

        if (dataPtr != NULL) {
            // Well then I just got a buffer back AND they must come back in the
            // order they were sent.  The next buffer to come back is _videoBufs[_curBuf].

            if ( dataPtr != _videoBufs[curBuf]) fprintf(stderr, "**** BUFFER ERROR \n");

            _bufStatus[curBuf] = 0;
            curBuf += 1;
            if (curBuf == _RVBUFS) curBuf = 0;

            //Check and see if this buffer went out in time.
            if ((oldmsc != -1) && (newmsc != oldmsc + _mscStep)) {
                fprintf(stderr, "MSC drop old = %lld, ", oldmsc);
                fprintf(stderr, "new = %lld\n", newmsc);
            }
            oldmsc = newmsc;
        }

        if (_bufStatus[playBuf] == 1) Sleep(5); 

    } while (_bufStatus[playBuf] == 1); 

    glReadPixels(0,0, _uSize, _vSize, GL_RGBA, GL_UNSIGNED_BYTE, _videoBufs[playBuf]);

    // Send the buffer to video out.
    p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
    p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)_videoBufs[playBuf];
    p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _videoSize;
    p[_VC_UST_INDEX].param = DM_VIDEO_UST;
    p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
    p[_VC_BUFFER_PV_END].param = DM_END;
    DC (dmPathSendBuffers(_hVideoPath, p));

    _bufStatus[playBuf] = 1;
    playBuf += 1;
    if (playBuf == _RVBUFS) playBuf = 0;

    /* Don't start the transfer until we have some fields queued up.        */
    if (count == 6) {
        fprintf(stderr, "Starting video transfer\n");
        DC (dmPathBeginTransfer(_hVideoPath) );
    }

    // Update gfx monitor depending on the rate specified. 
    if ( (_gfx_rate)  && ((count-1) % _gfx_rate == 0) ) {
        ZC ((*pwglMakeCurrentReadSGI)(_deviceContext, _PBdeviceContext, _renderContext));
        glCopyPixels(0, 0, _uSize, _vSize, GL_COLOR);
        if (_dbuffer) SwapBuffers(_deviceContext);
        ZC((*pwglMakeCurrentReadSGI)(_PBdeviceContext, _PBdeviceContext, _PBrenderContext));
    }

}

DWORD WINAPI ReadIt(LPVOID param)
{
    /*
     * This routine is used to implement a thread responsible for reading
     * samples from an individual file into a buffer in memory (one ReadIt()
     * thread per file).
     *
     * The "param" input parameter is a pointer to the stream_t structure
     * (see local variable "s") associated with the file.  Samples are read 
     * from disk into the circular buffer s->readvBufs at the entry pointed to 
     * by s->rvtail.  
     *
     * The effect thread consumes samples from the head of s->readvBufs, which
     * is pointed to by s->rvhead.  So, s->readvBufs is a circular buffer between 
     * the producer (ReadIt) and the consumer (PlayEfect). 
     * 
     * Remember - s->curField must be a multipe of s->iovcnt in order for this
     * read routine to remain nice and simple.  This is accomplished by forcing
     * s->f.Nfields to be a multiple of s->iovcnt (see InitStreamRead()).
     * This reader thread exits as soon as doit() sets "_IamPlaying" to 0.
     */

    stream_t *s;
    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;
    LONG hi_off, lo_off;
    DWORDLONG offset;

    s = (stream_t *)param;  /* Grab input buffer */

    offset = s->f.headerSize + (DWORDLONG)s->curField * s->f.readSize; 
    hi_off = (LONG)(offset>>32);
    lo_off = (LONG)(offset & 0x00000000ffffffff);
    SetFilePointer(s->f.fd, lo_off, &hi_off, FILE_BEGIN);

    while (_IamPlaying == 1) {
        fieldsToRead = ((s->curField + s->iovcnt) <= s->f.Nfields) ?
                         s->iovcnt : (s->f.Nfields - s->curField);

        /* Now make sure there is room in reader buffer to store 
         * "fieldsToRead" fields (which the consumer PutField() will later
         * display. 
         */
        do {
            full = 0;
            /* Check for the first full condition.  If the incremented
             * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
            tmphead = s->rvhead;
            tmptail = s->rvtail + fieldsToRead;
            if (tmptail == s->num_rvbufs) tmptail = 0;
            full = (tmptail == tmphead); 

            /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
            full |= ((tmphead > s->rvtail) && (tmphead < s->rvtail + fieldsToRead));

            if (full) {
                /* fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail); */ 
                Sleep(10);
            }
        } while (full);

        /* Read the fields.		*/
        tmptail = s->rvtail;
        ZC ( ReadFile(s->f.fd, s->readvBufs + tmptail*s->f.readSize, fieldsToRead*s->f.readSize, &i, NULL));
        tmptail += fieldsToRead;

        /* Inc. tail, indicating more vaild data is in buffer.	*/
        if (tmptail == s->num_rvbufs) tmptail = 0;
        s->rvtail = tmptail;
	
        /* Check for file wrap.		*/
        s->curField += fieldsToRead;

        if (s->curField >= s->f.Nfields) {
            s->curField = 0;
    	    SetFilePointer(s->f.fd, s->f.headerSize, NULL, FILE_BEGIN);
            fprintf(stderr, "Done pass through file\n");
        }

   }  /* end while(IamPLaying) */

   --_IamPlaying;
   s->rvtail = -1;

   ExitThread (-2);
}


void GetInputFields (char **f1, char **f2)
{
    /* This routine is called by the effect calculation
     * routines in common/fxcode.c.  It gives them pointers
     * to the next two uncompressed fields. The next
     * available field for a stream is at _s[i].rvhead (i=1,2).
     * If _s[i].rvhead == _s[i].rvtail then the queue is empty.
     */

    /* Wait if either of the reader buffers are empty.  */
    while  (_s[0].rvhead == _s[0].rvtail) {
        Sleep(5);
    }
    *f1 = _s[0].readvBufs + _s[0].rvhead*_s[0].f.readSize;

    if (_fx != FX_NONE) {
        while  (_s[1].rvhead == _s[1].rvtail) {
            Sleep(5);
        }
        *f2 = _s[1].readvBufs + _s[1].rvhead*_s[1].f.readSize;
    }
}



void GetInputArray (int nfields, char *fields[])
{
    int i;

    /* This routine is called by the effect calculation routines in
     * common/fxcode.c.  It returns pointers to the next pair of input
     * fields to process.
     */

    for (i=0; i < _nreaders; i++) {

        /* Wait if the reader buffer is empty.  */
        while  (_s[i].rvhead == _s[i].rvtail) {
            Sleep(5);
        }

        fields[i] = _s[i].readvBufs + _s[i].rvhead*_s[i].f.readSize;
    }
}



void Quit(HWND wnd)
{
    int i;

    /* Stop play */
    if (_IamPlaying) doit();

    ZC (wglMakeCurrent(NULL, NULL));
    ZC ((*pwglMakeCurrentReadSGI)(NULL, NULL, NULL));
	ZC (wglMakeCurrent(_deviceContext, _renderContext));
	ZC (wglDeleteContext(_renderContext));
	ZC (wglMakeCurrent(_deviceContext, _PBrenderContext));
	ZC (wglDeleteContext(_PBrenderContext));

	ReleaseDC(wnd, _deviceContext);

    for (i=0; i < _nstreams; i++) {
        CloseHandle(_s[i].f.fd);
    }

	PostQuitMessage(0);
}



void initVideo(dmJackTypeEnum jacktype)
{
    int i, numbufs;
    int format, capType;
    int videoHeight;
    DMpv *sysCaps, *jackCaps, *pathCaps;
    DMint64 *pJacks, *pPaths;
    DMint32 nJacks, nPaths;
    DMpv p[10];

    DC (dmSystemGetCapabilities(DM_SYSTEM_LOCALHOST, &sysCaps));
    pJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].value.pInt64;
    nJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].length;

    // Find a jack of the given type, i.e. composite, digial, etc.).
    for (i = 0; i < nJacks; i++) {
        DC (dmJackGetCapabilities(pJacks[i], &jackCaps));

        if (jackCaps[DM_JACK_DIRECTION_INDEX].value.int32 == DM_DIRECTION_OUT &&
            jackCaps[DM_JACK_TYPE_INDEX].value.int32 == jacktype){
            break;
        }
    }

    ERR ((i == nJacks), "initVideo: can't find a jack");

    // For the given jack, find the dev to mem path.
    pPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].value.pInt64;
    nPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].length;
    for (i = 0; i < nPaths; i++) {
        DC (dmPathGetCapabilities(pPaths[i], &pathCaps));
        if (pathCaps[DM_PATH_TYPE_INDEX].value.int32 == DM_PATH_TYPE_MEM_TO_DEV) break;
    }

    ERR ((i == nPaths), "initVideo: can't find a path");

    DC (dmPathOpen(pPaths[i], NULL, &_hVideoPath));


    format = DM_FORMAT_RGBA_8_FULL;

    if (_vSize < 400) {
        capType =  DM_INTERLEAVE_MODE_SINGLE_FIELD;
        _mscStep = 1;
        videoHeight = _vSize;
    }
    else {
        capType = DM_INTERLEAVE_MODE_INTERLEAVED;
        _mscStep = 2;
        videoHeight = _vSize / 2;
    }

    fprintf(stderr, "Video height = %d, sample height = %d\n", videoHeight, _vSize);

    //Set the image format and interleave mode on the path.
    i = 0;
    dmPvSetInt32Param(p[0], DM_VIDEO_WIDTH_INT32, _uSize);
    dmPvSetInt32Param(p[1], DM_VIDEO_HEIGHT_F1_INT32, videoHeight);
    dmPvSetInt32Param(p[2], DM_VIDEO_HEIGHT_F2_INT32, videoHeight);
    dmPvSetInt32Param(p[3], DM_IMAGE_FORMAT_INT32, format);
    dmPvSetInt32Param(p[4], DM_IMAGE_INTERLEAVE_MODE_INT32, capType);	
    dmPvSetInt32Param(p[5], DM_END, 0);
    DC (dmPathSendControls(_hVideoPath, p));

}

int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc)
{
    // Routine to check message queue attached to video hardware.
    // In particular, this routine gets the address, msc, and ust of
    // buffers that have been processed by the video hardware (i.e. sent
    // to video output.

    DMint32 message;
    DMpv *reply;
    DMstatus ret;
    int rval = 0;
 
    *data = NULL;
    *ust =  0;
    *msc =  0;

    ret = dmPathReceiveMessage(_hVideoPath, &message, &reply);

    if (ret != DM_STATUS_NO_ERROR) {     //Did not get error free result.

		if (ret != DM_STATUS_RECEIVE_QUEUE_EMPTY) {  //Did not get empty queue
	    	fprintf(stderr, "dmPathReceiveMessage returned value %x\n", ret);
		}
		return -1;  //Either error or queue empty.
    }

    switch (message) {
    case DM_MESSAGE_TRANSFER_COMPLETE:
		*data = (unsigned char *)reply[_VC_BUFFER_INDEX].value.pByte;
		*ust = reply[_VC_UST_INDEX].value.int64;
		*msc = reply[_VC_MSC_INDEX].value.int64;
		break;
    case DM_MESSAGE_MISSED_FIELD:
        //fprintf(stderr, "DM_MESSAGE_MISSED_FIELD\n");
		break;
    case DM_MESSAGE_TRANSFER_ABORTED:
		*data = (unsigned char *)reply[_VC_BUFFER_INDEX].value.pByte;
        //fprintf(stderr, "DM_MESSAGE_TRANSFER_ABORTED\n");
		break;
	case DM_MESSAGE_CONTROLS_SUCCESSFUL:
        fprintf(stderr, "DM_MESSAGE_CONTROLS_SUCCESSFUL\n");
		break;
    default:
        //fprintf(stderr, "dmPathReceiveMessage returned message %x\n", ret);
		break;
    }

    return rval;
}


void initStreamRead(int iovcnt, stream_t *s)
{
    int i, n;

    s->curField = 0;
    s->iovcnt = iovcnt;

    /* Cheat a liitle here, it makes the reading a lot simpler
     * if both files contain a multiple of iovcnt fields.
     */
    if (s->f.Nfields < s->iovcnt) {
        s->iovcnt = s->f.Nfields;
    }
    else if ((s->f.Nfields % s->iovcnt) != 0) {
        s->f.Nfields -= (s->f.Nfields % s->iovcnt);
    }

    /* Set up the buffers that the readers will place the uncompressed
     * fields into.
     */

    s->num_rvbufs = 3*s->iovcnt;
    n = s->num_rvbufs * s->f.readSize;

    /* Allocate aligned memory. */
    NC (s->readvBufs = (char * )VirtualAlloc(NULL, n, MEM_COMMIT, PAGE_READWRITE));
}


int GetNumReaders(FxType_t fx)
{
    if (fx == FX_NONE) return 1;
    else if ((fx == FX_3STREAM_WIPE) || (fx == FX_3STREAM_TEX)) return 3;
    else if (fx == FX_4STREAM) return 4;
    else return 2;
}
