/**************************************************
   solar.c
   
   Pressing the "a" key toggles the animation
   Pressing the up and down arrow keys will
   increase/decrease the animation rate

**************************************************/

#include "glos.h" // MS specifc stuff

#include <GL/gl.h> // system OpenGL includes
#include <GL/glu.h>
#include <GL/glaux.h>

static GLenum spinMode = GL_TRUE;

void OpenGLInit(void);

static void CALLBACK Animate(void );
static void CALLBACK Key_a(void );
static void CALLBACK Key_up(void );
static void CALLBACK Key_down(void );
static void CALLBACK ResizeWindow(GLsizei w, GLsizei h);

static int HourOfDay = 0, DayOfYear = 0;
static int AnimateIncrement = 24; // in hours


static void CALLBACK Key_a(void)
{
    spinMode = !spinMode;
}

static void CALLBACK Key_up(void)
{
    AnimateIncrement *= 2;
	if ( 0 == AnimateIncrement )
		AnimateIncrement = 1;
}

static void CALLBACK Key_down(void)
{
    AnimateIncrement /= 2;
	
}

static void CALLBACK Animate(void)
{
	float spot_ambient[] = {1.0, 1.0, 1.0, 1.0};
	float spot_position[] = {0.0, 1.0, 0.0, 0.0};

    float local_view[] = {-10.0};

	float ambient1[] = {1.0, 0.0, 0.0};
	float ambient2[] = {0.0, 0.0, 1.0};
	float ambient3[] = {1.0, 1.0, 0.0};

    float diffuse[] = {0.61, 0.04, 0.04};
    float specular[] = {0.72, 0.62, 0.62};


	// clear the redering window
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLightfv(GL_LIGHT0, GL_POSITION, spot_position);

	glLightfv(GL_LIGHT0, GL_AMBIENT, spot_ambient);

    glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, local_view);

    glFrontFace(GL_CW);

    glEnable(GL_LIGHTING);

    glEnable(GL_LIGHT0);

    glEnable(GL_AUTO_NORMAL);
    glEnable(GL_NORMALIZE);

    glMaterialfv(GL_FRONT, GL_AMBIENT, ambient1);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
    glMaterialf(GL_FRONT, GL_SHININESS, 0.6*128.0);

    if (spinMode)
		{
		// calc animation parameters
        HourOfDay += AnimateIncrement;
        DayOfYear += HourOfDay/24;

        HourOfDay = HourOfDay%24;
        DayOfYear = DayOfYear%365;
		}
	
	// clear current matrix (Modelview)
    glLoadIdentity();
	// back off five units 
    glTranslatef ( 0.0f, 0.0f, -5.0f );
	// rotate the plane of the elliptic
	// (rotate the model's plane about the
	// x axis by five degrees)
	glRotatef( 5.0f, 1.0f, 0.0f, 0.0f );

    // draw the sun
	auxSolidSphere(1.0);

    // draw the Earth
    glRotatef( (GLfloat)(360.0*DayOfYear/365.0),
		0.0, 1.0, 0.0 );
    glTranslatef( 4.0, 0.0, 0.0 );
    glPushMatrix(); // save matrix state
	glRotatef( (GLfloat)(360.0*HourOfDay/24.0) 
		  , 0.0, 1.0, 0.0 );

	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient2);

	auxSolidSphere( 0.2 );

	//draw the moon

    glPopMatrix(); // restore matrix state

   	glRotatef( (GLfloat)(360.0*12.5*DayOfYear/365.0),
		0.0, 1.0, 0.0 );
    glTranslatef( 0.5, 0.0, 0.0 );

	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient3);

	auxSolidSphere( 0.05);

	// flush the pipeline, swap the buffers
    glFlush();
    auxSwapBuffers();

}

// initialize OpenGL
void OpenGLInit(void)
{
	

    glShadeModel( GL_FLAT );
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
    glClearDepth( 1.0f );
	glDepthFunc( GL_LEQUAL );
    glEnable( GL_DEPTH_TEST );
	glShadeModel(GL_SMOOTH);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    

}

// called when the window is resized
static void CALLBACK ResizeWindow(GLsizei w, GLsizei h)
{
    h = (h == 0) ? 1 : h;
	w = (w == 0) ? 1 : w;
	glViewport( 0, 0, w, h );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

	glOrtho( -10.0, 10.0, -5.0, 5.0, 0.5, 10.0 );

	// select the Modelview matrix
    glMatrixMode( GL_MODELVIEW );
}


// main routine
// set up OpenGL, hook up callbacks
// and start the main loop
int main( int argc, char** argv )
{
	// we're going to animate it, so double buffer 
    auxInitDisplayMode(AUX_DOUBLE | AUX_RGB );
    auxInitPosition( 0, 0, 600, 330 );
    auxInitWindow( "Solar System" );

	// Initialize OpenGL as we like it..
    OpenGLInit();

	// set up callback functions
    auxKeyFunc( AUX_UP, Key_up ); // faster
    auxKeyFunc( AUX_DOWN, Key_down ); // slower
	auxKeyFunc( AUX_a, Key_a ); //animate
    auxReshapeFunc( ResizeWindow );

	// call this when idle
    auxIdleFunc( Animate );
	// call this in main loop
	auxMainLoop( Animate );

    return(0);
}
