/*
 * Copyright (c) 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <iostream.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>

/////////////////////////////////////////////////////////////////////////////
//
// function prototypes
//
/////////////////////////////////////////////////////////////////////////////

bool hasExtension( const char * extname );
void init();
void draw();
void timed_draw();
void draw_quads( bool do_hint );
void make_quads();
void idle();
void reshape( int ww, int hh );
void key( unsigned char key, int x, int y );

static const int _num_quads = 100000;
static const int _num_verts = _num_quads*4;
static float _qq[ _num_verts ][ 3 ];

/* opengl v1.1 defines */

#define GL_CLIP_VOLUME_CLIPPING_HINT_EXT    0x80F0

/////////////////////////////////////////////////////////////////////////////
//
// ogl initialization
//
/////////////////////////////////////////////////////////////////////////////

static void init()
{
  if ( false == hasExtension( "GL_EXT_clip_volume_hint" ) )
    {
      printf( "system doesn't support clip volume hint\n" );
      exit( -1 );
    }

  glDrawBuffer( GL_FRONT );
  glClearColor( 0, 0, 0, 0 );

  glDisable( GL_DEPTH_TEST );
  make_quads();
}

bool hasExtension( const char * ext )
{
  bool result = false;
  char *exts = (char *) glGetString( GL_EXTENSIONS );

  if ( exts != NULL )
    {
      if ( NULL != strstr( exts, ext ) )
	{
	  result = true;
	}
    }

  return( result );
}

/////////////////////////////////////////////////////////////////////////////
//
// drawing / resizing functions
//
/////////////////////////////////////////////////////////////////////////////

static void reshape( int ww, int hh )
{
  glViewport( 0, 0, ww, hh );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  glOrtho( 0.0, 1.0, 0.0, 1.0, -10.0, 10.0 );

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
}

static void draw()
{
  glClear( GL_COLOR_BUFFER_BIT );

  glBegin( GL_TRIANGLES );

  glColor3f( 1.0, 1.0, 1.0 );
  glVertex3f( 0.0, 0.0, 0.0 );
  glVertex3f( 1.0, 0.0, 0.0 );
  glVertex3f( 0.0, 1.0, 0.0 );

  glEnd();
}

static void timed_draw()
{
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  int ii;
  clock_t start;
  clock_t stop;

  double time_hint;
  double time_no_hint;

  int ntimes = 100;

  /* draw w/GL_EXT_clip_volume_hint enabled */
  start = clock(); 
  for ( ii=0; ii<ntimes; ii++ )
    {
      draw_quads( true );
    }
  stop = clock();
  time_hint = (double)( stop - start ) / CLOCKS_PER_SEC;

  printf( "using: GL_EXT_clip_volume_hint time (s):    %f\n", time_hint );

  /* draw w/GL_EXT_clip_volume_hint disabled */
  start = clock();
  for ( ii=0; ii<ntimes; ii++ )
    {
      draw_quads( false );
    }
  stop = clock();
  time_no_hint = (double)( stop - start) / CLOCKS_PER_SEC;

  printf( "!using: GL_EXT_clip_volume_hint time (s):   %f\n", time_no_hint );

  printf( "%% performance improvement: %f\n",
	 (time_no_hint-time_hint)*100.0/time_no_hint );
  printf( "\n" );
}

static void make_quads()
{
  srand( ( unsigned int ) time( NULL ) );

  float size = float( .01 );
  
  float offset = size * 2;
  float range  = 1.0 - 2 * offset;

  for ( int ii=0; ii<_num_quads*4; ii++ )
   {
     float center[ 2 ];

     center[ 0 ] = ( range * float(  rand() )/float( RAND_MAX ) ) + offset;
     center[ 1 ] = ( range * float(  rand() )/float( RAND_MAX ) ) + offset;
     
     // ll, lr, ur, ul
     
     _qq[ ii ][ 0 ] = center[ 0 ] - size;
     _qq[ ii ][ 1 ] = center[ 1 ] - size;
     _qq[ ii ][ 2 ] = 0;
     
     ii++;
     
     _qq[ ii ][ 0 ] = center[ 0 ] + size;
     _qq[ ii ][ 1 ] = center[ 1 ] - size;
     _qq[ ii ][ 2 ] = 0;
     
     ii++;
     
     _qq[ ii ][ 0 ] = center[ 0 ] + size; 
     _qq[ ii ][ 1 ] = center[ 1 ] + size; 
     _qq[ ii ][ 2 ] = 0;
     
     ii++;;
     
     _qq[ ii ][ 0 ] = center[ 0 ] - size; 
     _qq[ ii ][ 1 ] = center[ 1 ] + size; 
     _qq[ ii ][ 2 ] = 0;
   }

}

static void draw_quads( bool do_hint )
{
  /* check if we should enable hint */
  if ( true == do_hint )
    {
      /* enable == GL_FASTEST */
      glHint( GL_CLIP_VOLUME_CLIPPING_HINT_EXT, GL_FASTEST );
    }
  else
    {
      /* disable == GL_NICEST */
      glHint( GL_CLIP_VOLUME_CLIPPING_HINT_EXT, GL_NICEST );
    }

  /* draw the quads */

  glBegin( GL_QUADS );
  int ii;
  for ( ii=0; ii<_num_quads; ii++ )
    {
      glColor3f(  0.0, 0.0, (float)ii/_num_quads );
      for ( int jj=0; jj<4; jj++ )
	{
	  glVertex3fv( _qq[ ii*4 + jj ] );
	}
    }
  glEnd();
  
}

/////////////////////////////////////////////////////////////////////////////
//
// key-event handler
//
/////////////////////////////////////////////////////////////////////////////

static void key( unsigned char key, int xx, int yy )
{
  switch( key )
    {
      
    case 't':
      {
	timed_draw();
      }
      break;
      
    case 'q':
    case 27:
      {
	exit(0);
      }
    }
}

/////////////////////////////////////////////////////////////////////////////
//
// main (setup, visual selection, callback registration)
//
/////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  glutInit( &argc, argv );
  
  glutInitDisplayMode( GLUT_RGB | GLUT_SINGLE );
  glutInitWindowSize( 500, 500 );
  glutCreateWindow( "clip_volume_test" );
  
  init();
  
  glutReshapeFunc( reshape );
  glutKeyboardFunc( key );
  glutDisplayFunc( draw );
  glutMainLoop();

  return( 1 );
}
