/*
 * Copyright (c) 1997, Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "sglock.h"

#ifdef DEBUG
void CSgLock::timeout()
{
  struct timeval tim;
  gettimeofday(&tim);
  printf(" sec:%d usec:%d\n", tim.tv_sec, tim.tv_usec);
}
#endif

void CSgLock::init()
{
  static struct sembuf args[2*MAX_CLIENTS];

  for (int i = 0; i < MAX_CLIENTS; i++) {
    args[(2*i)].sem_num = (2*i);
    args[(2*i)+SEM_SERVER].sem_num = (2*i)+SEM_SERVER;
    args[(2*i)].sem_op = 1;
    args[(2*i)+SEM_SERVER].sem_op = 1;
  }
  if (semop(lockId, args, 2*MAX_CLIENTS) == -1)
    perror("semop");
}

CSgLock::CSgLock(int key, int server)
{
  // try to get an existing semaphore first
  if (server) {
    if ((lockId = semget(key, 0, 0)) != -1) // key exists
      remove();
  }
  if ((lockId = semget(key, 0, 0)) == -1) {
    if (ENOENT == errno) {
      lockId = semget(key, 2*MAX_CLIENTS, IPC_CREAT | 0666);
#ifdef DEBUG
       if (lockId == -1)
         perror("semget");
#endif
      // initialize
      this->init();
    }
#ifdef DEBUG
    else
      perror("semget");
#endif
  }
}

// this shouldn't block since it should be set to zero after a write
void CSgLock::client_read(int client)
{
  static struct sembuf args;

#ifdef DEBUG
  printf("client attempting to read");
  this->timeout();
#endif
  // client reads from server section
  if (-1 == client)
    args.sem_num = SEM_SERVER;
  else
    args.sem_num = (2*client)+SEM_SERVER+2;
  args.sem_op = 0;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("client is reading S[0]");
  this->timeout();
#endif

}

// this resets the value to 1 after a read
void CSgLock::client_finish_read(int client)
{
  static struct sembuf args;

#ifdef DEBUG
  printf("client is attempting to finish reading");
  this->timeout();
#endif
  // client needs to unlock server section so it can be written to
  if (-1 == client)
    args.sem_num = SEM_SERVER;
  else
    args.sem_num = (2*client)+SEM_SERVER+2;
  args.sem_op = 1;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("client is finished reading S[1]");
  this->timeout();
#endif

}

// this should set it to zero for a read
void CSgLock::client_write(int client)
{
  static struct sembuf args;
#ifdef DEBUG
  printf("client is attempting to write");
  this->timeout();
#endif
  // client writes to client section
  if (-1 == client)
    args.sem_num = SEM_CLIENT;
  else
  args.sem_num = (2*client)+SEM_CLIENT+2;
  args.sem_op = -1;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("client is writing C[0]");
  this->timeout();
#endif

}

// this should set it to 0 and then immediately unset it
int CSgLock::client_verify_write(int client)
{
  int semret;
  static struct sembuf args;
#ifdef DEBUG
  printf("client verifying write");
  this->timeout();
#endif
  // client waits for server to read the client section
  if (-1 == client)
    args.sem_num = SEM_CLIENT;
  else
    args.sem_num = (2*client)+SEM_CLIENT+2;
  args.sem_op = -1;
  args.sem_flg = IPC_NOWAIT;
  semret = semop(lockId, &args, 1); // set it to 0
  if (-1 == semret)
    return 0;
#ifdef DEBUG
  printf("client partial verify C[0]");
  this->timeout();
#endif
  args.sem_op = 1;
  semret = semop(lockId, &args, 1); // set it back to 1
  if (-1 == semret)
    return -1;  // grave error has occurred
#ifdef DEBUG
  printf("client write verified C[1]");
  this->timeout();
#endif
  return 1;
}

// this shouldn't block since it should be set to zero after a write
void CSgLock::server_read(int client)
{
  static struct sembuf args;
#ifdef DEBUG
  printf("server is attempting to read");
  this->timeout();
#endif
  // server reads from client section
  if (-1 == client)
    args.sem_num = SEM_CLIENT;
  else
    args.sem_num = (2*client)+SEM_CLIENT+2;
  args.sem_op = 0;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("server is reading C[0]");
  this->timeout();
#endif

}

// this resets the value to 1 after a read
void CSgLock::server_finish_read(int client)
{
  static struct sembuf args;

#ifdef DEBUG
  printf("server is attempting to finish reading");
  this->timeout();
#endif
  // server needs to reset client section for writing
  if (-1 == client)
    args.sem_num = SEM_CLIENT;
  else
    args.sem_num = (2*client)+SEM_CLIENT+2;
  args.sem_op = 1;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("server is finished reading C[1]");
  this->timeout();
#endif

}

// this should set it to zero for a read
void CSgLock::server_write(int client)
{

  static struct sembuf args;
#ifdef DEBUG
  printf("server is attempting to write");
  this->timeout();
#endif
  // server writes to server section
  if (-1 == client)
    args.sem_num = SEM_SERVER;
  else
    args.sem_num = (2*client)+SEM_SERVER+2;
  args.sem_op = -1;
  semop(lockId, &args, 1);

#ifdef DEBUG
  printf("server is writing S[0]");
  this->timeout();
#endif

}

// this should set it to 0 and then immediately unset it
int CSgLock::server_verify_write(int client)
{
  int semret;

  static struct sembuf args;
#ifdef DEBUG
  printf("server verifying write");
  this->timeout();
#endif
  // server waits for client to read the server section
  if (-1 == client)
    args.sem_num = SEM_SERVER;
  else
    args.sem_num = (2*client)+SEM_SERVER+2;
  args.sem_op = -1;
  args.sem_flg = IPC_NOWAIT;
  semret = semop(lockId, &args, 1); // set it to 0
  if (semret == -1)
    return 0;
#ifdef DEBUG
  printf("server partial verify S[0]");
  this->timeout();
#endif
  args.sem_op = 1;
  semret = semop(lockId, &args, 1); // set it back to 1
  if (semret == -1)
    return -1;  // an error has occurred
#ifdef DEBUG
  printf("server write verified S[1]");
  this->timeout();
#endif
  return 1;
}

void CSgLock::remove()
{
  semctl(lockId, 0, IPC_RMID, 0);
}
