/*
 * Copyright (c) 1997, Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "sgmessage.h"

int CSgMsg::constr_name(struct sockaddr *addr, const char *pathnm)
{
  addr->sa_family = AF_UNIX;
  strcpy(addr->sa_data, pathnm);

  return sizeof(addr->sa_family) + strlen(pathnm) + 1;

};

int CSgMsg::setup_callbacks(int id)
{

  if (-1 == id)
    return -1;

  // setup the input callbacks to trigger on read and except
  if (id == listen_sid) { // setup on the listening socket
    l_readId = ::XtAppAddInput(*app, id, (XtPointer)XtInputReadMask, proc,
                            this);
    l_exceptId = ::XtAppAddInput(*app, id, (XtPointer)XtInputExceptMask, proc,
                              this);
    if ((0 == l_readId) || (0 == l_exceptId)) {
      // problem with the add input stuff
      if (0 != l_readId)
       ::XtRemoveInput(l_readId);
      if (0 != l_exceptId)
       ::XtRemoveInput(l_exceptId);
      return -1;
    }
    else
      return 0;
  }
  else if (id == sid) { // setup on the connected socket
    readId = ::XtAppAddInput(*app, id, (XtPointer)XtInputReadMask, proc,
                            this);
    exceptId = ::XtAppAddInput(*app, id, (XtPointer)XtInputExceptMask, proc,
                              this);
    if ((0 == readId) || (0 == exceptId)) {
      // problem with the add input stuff
      if (0 != readId)
       ::XtRemoveInput(readId);
      if (0 != exceptId)
       ::XtRemoveInput(exceptId);
      return -1;
    }
    else
      return 0;
  }
  else // unknown id being passed in
    return -1;

}

int CSgMsg::remove_callbacks(int id)
{

  if (-1 == id)
    return -1;

  if (id == listen_sid) { // this is the listening socket
    if (0 != l_readId) {
      ::XtRemoveInput(l_readId);
      l_readId = 0;
    }
    if (0 != l_exceptId) {
      ::XtRemoveInput(l_exceptId);
      l_exceptId = 0;
    }
    return 0;
  }
  else if (id == sid) { // this is the connected id
    if (0 != readId) {
      ::XtRemoveInput(readId);
      readId = 0;
    }
    if (0 != exceptId) {
      ::XtRemoveInput(exceptId);
      exceptId = 0;
    }
    return 0;
  }
  else  // unknown id
    return -1;
}

int CSgMsg::listen()
{
  // we should only be listening on the listen socket
  int retVal;

  retVal = ::listen(listen_sid, 5);

  return retVal;
}

void CSgMsg::check_file(const char *pathnm)
{
  struct stat statbuf;

  // Check for the existence of the socket file
  if (::stat(pathnm, &statbuf) < 0) {
    if (errno == ENOENT) // the file doesn't exist
      return;
#ifdef DEBUG
    else
      perror("stat");
#endif
  }

  // Check if the file is an old socket
  if (S_ISSOCK(statbuf.st_mode)) {
#ifdef DEBUG
    int retVal;
    // remove it
    retVal = ::unlink(pathnm);
    if (retVal < 0)
      perror("unlink");
#else
    ::unlink(pathnm);
#endif
  }
  else
    fprintf(stderr, "SOCKET file %s already exists as a normal file.\nPlease remove it.\n", pathnm);
}

CSgMsg::CSgMsg(char *pathnm, XtAppContext *context,
                  XtInputCallbackProc call) :
  listen_sid(-1), sid(-1), app(context), readId(0),
  exceptId(0), proc(call), l_readId(0), l_exceptId(0),
  data_ready(NULL)
{
  sid = ::socket(AF_UNIX, SOCK_STREAM, 0);
#ifdef DEBUG
  if (sid < 0)
    perror("socket");
#endif
  strcpy(name, pathnm);
}

CSgMsg::CSgMsg(char *pathnm, void (*data_func)(CSgMsg *,int))
  : listen_sid(-1), sid(-1), app(NULL), readId(0),
    exceptId(0), proc(NULL), l_readId(0), l_exceptId(0),
    data_ready(data_func)
{
  sid = ::socket(AF_UNIX, SOCK_STREAM, 0);
#ifdef DEBUG
  if (sid < 0)
    perror("socket");
#endif
  strcpy(name, pathnm);
}

CSgMsg::CSgMsg(char *pathnm, int fid) : listen_sid(-1),
  sid(fid), app(NULL), readId(0), exceptId(0), proc(NULL),
  l_readId(0), l_exceptId(0), data_ready(NULL)
{
  strcpy(name, pathnm);
}

CSgMsg::~CSgMsg()
{

  // cleanup time
  if (-1 != sid)
    this->close();

  if (-1 != listen_sid) {
    this->remove_callbacks(listen_sid);
    ::close(listen_sid);
    listen_sid = -1;
    // remove the unix file created for communication
    ::unlink(name);
  }
}

int CSgMsg::bind()
{
  // If we are binding then it is the listening socket
  listen_sid = sid;
  sid = -1;

  struct sockaddr addr;
  int len = constr_name(&addr, name);
  int retVal;

  // check if file already exists and remove it if it does
  check_file(name);

  retVal = ::bind(listen_sid, &addr, len);
  if (retVal < 0) {
#ifdef DEBUG
    perror("bind");
#endif
    return retVal;
  }

  // set up a listen on the listening socket
  retVal = this->listen();
  if (retVal < 0) {
#ifdef DEBUG
    perror("listen");
#endif
    return retVal;
  }

  // set up the input callbacks to trigger on read or write events
  if ( proc != NULL )
    retVal = this->setup_callbacks(listen_sid);
#ifdef DEBUG
  if (retVal < 0)
    fprintf(stderr, "Problem setting up callbacks");
#endif

  return retVal;
}

int CSgMsg::connect()
{

  // connect should only be done by the client
  int retVal;

  if (sid < 0)
    return -1;

  struct sockaddr addr;
  int len = this->constr_name(&addr, name);
  retVal = ::connect(sid, &addr, len);
  if (retVal < 0) {
#ifdef DEBUG
    perror("connect");
#endif
    return retVal;
  }

  // set up the input callbacks to trigger on read and write events
  retVal = this->setup_callbacks(sid);
#ifdef DEBUG
  if (retVal < 0)
    fprintf(stderr, "Problem setting up the callbacks\n");
#endif
  return retVal;
}

CSgMsg * CSgMsg::accept()
{
  // accept always occurs on the listening socket

  if (listen_sid < 0)
    return 0;

  sid = ::accept(listen_sid, 0, 0);
  if (sid < 0) {
#ifdef DEBUG
    perror("accept");
#endif
    return 0;
  }

  CSgMsg *retMsg = new CSgMsg(name, sid);
  // set up the input callbacks to trigger on read or write events
  //retMsg->setup_callbacks(sid);

  this->sid = -1;

  return retMsg;

}

int CSgMsg::read(void *buf, int len)
{
  int retVal;

  retVal = ::read(sid, buf, len);
#ifdef DEBUG
  if (retVal < 0)
    perror("read");
#endif

  return retVal;

}

int CSgMsg::write(void *buf, int len)
{

  int retVal;

  retVal = ::write(sid, buf, len);
#ifdef DEBUG
  if (retVal < 0)
    perror("write");
#endif

  return retVal;

}

int CSgMsg::close()
{
  // close an opened socket connection the listen socket is
  // destroyed by deleting the object
  int retVal;

  if (-1 == sid) // closeing a non-open socket
    return -1;

  retVal = this->remove_callbacks(sid);
  if (retVal < 0) {
#ifdef DEBUG
    fprintf(stderr, "Problem removing callbacks\n");
#endif
    return retVal;
  }

  retVal = ::close(sid);
#ifdef DEBUG
  if (retVal < 0)
    perror("close");
#endif
  if (retVal >= 0)
    sid = -1;

  return retVal;

}

// return 1 if it's a connect else return 0
int CSgMsg::is_connect(int fid, XtInputId xid)
{
  int retVal;

  if ((fid == listen_sid) && (xid == l_readId)) {
    int nbytes;
    retVal = ioctl(fid, FIONREAD, &nbytes);
    if (retVal != 0) {
#ifdef DEBUG
      perror("ioctl");
#endif
      return retVal;
    }
    if (0 == nbytes)
      return 1;
    else
      return 0;
  }
  else // can't be a connect on a non listening socket
    return 0;
}

// return 1 if it's a read event 0 otherwise
int CSgMsg::is_read(int fid, XtInputId xid)
{
  int retVal;

  if (fid == listen_sid) {
    if (xid == l_readId) {
      int nbytes;
      retVal = ioctl(fid, FIONREAD, &nbytes);
      if (retVal != 0) {
#ifdef DEBUG
       perror("ioctl");
#endif
       return retVal;
      }
      if (0 != nbytes) // if there is something to read
       return 1;
      else
       return 0;
    }
    else
      return 0;
  }
  else if (fid == sid) {
    if (xid == readId) {
      int nbytes;
      retVal = ioctl(fid, FIONREAD, &nbytes);
      if (retVal != 0) {
#ifdef DEBUG
       perror("ioctl");
#endif
       return retVal;
      }
      if (0 != nbytes) // if there is something to read
       return 1;
      else
       return 0;
    }
    else
      return 0;
  }
  else // don't know who this is from
    return 0;
}

// return 1 if it's an except event 0 otherwise
int CSgMsg::is_except(int fid, XtInputId xid)
{
  if (fid == listen_sid) {
    if (xid == l_exceptId)
      return 1;
    else
      return 0;
  }
  else if (fid == sid) {
    if (xid == exceptId)
      return 1;
    else
      return 0;
  }
  else // don't know who this is from
    return 0;
}

// return 1 if it's a close event 0 otherwise
int CSgMsg::is_close(int fid, XtInputId xid)
{
  int retVal;

  if (fid == listen_sid) // we can't get a close event on the listen
    return 0;
  else if ((fid == sid) &&(xid == readId)) {
    int nbytes;
    retVal = ioctl(fid, FIONREAD, &nbytes);
    if (retVal != 0) {
#ifdef DEBUG
      perror("ioctl");
#endif
      return retVal;
    }
    if (0 == nbytes)
      return 1;
    else
      return 0;
  }
  else
    return 0;
}

int CSgMsg::check_for_messages()
{
  fd_set readfds;
  fd_set exceptfds;
  struct timeval timeout;
  int can_read;
  int max;

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;

  FD_ZERO(&readfds);
  FD_ZERO(&exceptfds);
  if ( 0 != sid ) {
    FD_SET(sid, &readfds);
    FD_SET(sid, &exceptfds);
  }
  if ( 0 != listen_sid ) {
    FD_SET(listen_sid, &readfds);
    FD_SET(listen_sid, &exceptfds);
  }

  max = (sid > listen_sid) ? sid : listen_sid;

  can_read = select( max+1, &readfds, NULL, &exceptfds,
                    &timeout );
  if ( -1 == can_read ) {
    if (EINTR != errno)
      fprintf(stderr, "Select problem\n");
    return -1;
  }
  else if ( 0 == can_read ) // we timed out
    return 0;
  else if ( can_read > 0 ) { // something is set
    if (FD_ISSET(sid, &exceptfds) || FD_ISSET(listen_sid,
                                             &exceptfds)) {
      (data_ready)(this, SGMSG_EXCEPT);
      return 1;
    }
    else if (FD_ISSET(listen_sid, &readfds)) {
      (data_ready)(this, SGMSG_CONNECT);
      return 1;
    }
    else if (FD_ISSET(sid, &readfds)) {
      int nbytes;
      if ( 0 != ioctl(sid, FIONREAD, &nbytes) ) {
#ifdef DEBUG
       perror("ioctl");
#endif
       return -1;
      }
      if ( 0 == nbytes ) {
       (data_ready)(this, SGMSG_CLOSE);
       return 1;
      }
      else {
       (data_ready)(this, SGMSG_READ);
       return 1;
      }
    }
  }

  return 0;
}
