/*
 * Copyright (c) 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* tplay_u  = Plays uncompressed video fields or frames from disk to the 
 *		      gfx monitor using texture memory to zoom or shrink the image.
 *            Either the OpenGL color matirx functionality OR a software 
 *		      algorithm can be used to convert the YUV data to RGB before 
 *		      loading it into texture memory.
 *
 *            The choices for data format are:
 *               VL_PACKING_ABGR_8_P = GL_RGBA
 *               VL_PACKING_BGR_8_P  = GL_RGB (denoted rgbp).
 *               VL_PACKING_UYV_8_P  = 24 bit packed yuv pixels (denoted yuvp).
 *               VL_PACKING_YVYU_422_8 = 16 bit yuv 4:2:2 sampled pixels.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength> 	- Number of reads per read call.
 *     <-gfx winx winy>    	- Size of gfx display window (texture zoom size). 
 *     <-single>		    - Force single buffering (default is double).
 *     <-cc sw|cm|dir>	    - Specifies how to color convert yuv data:
 *                             -cc sw  = do software conversion (422_8 only). 
 *                             -cc cm  = do conversion with color matirx (422_8, yuvp). 
 *                             -cc dir = Send yuv data directly to OpenGL (uses the 
 *                                       YCRCB_SGIX extensions. 
 *				              of color matrix hardware (YUV 4:2:2 data only).
 *     <-inter>			    - Interlace fields into frames (4:2:2 data with -cc sw|cm).
 *     fileName			    - file created with vgrab_u. 
 *
 *
 * When you run the program, a blank window appears.  Click once to start
 * play, click again to stop play, etc. This program has been run on an 
 * Octane MXI.
 *
 * There are three threads sprco'd off in this program and they communicate 
 * through two buffer structres, _b[READER] (the reader queue) and _b[CONVERTER]
 * (the converter queue).  New items are ALWAYS added to the tail of a queue and
 * old items are always removed from the head of a queue.  Here are the threads:
 *
 *	1. ReadIt()   - This thread uses ReadFile() to grab fields/frames of data
 *	     	        from disk. 
 *		       
 *	                This thread reads the fields into the "tail" of the reader
 *			        queue (_b[READER].bufs points to start of queue). The data 
 *			        then flows to:
 *
 *              					    | (do step 2 if YUV 4:2:2 with -cc sw|cm,
 *              					    |     otherwise go to step 3.)
 *		                		      \ | /
 *	                			       \|/
 *		                			    V
 *
 *	2. Expander() - This thread grabs data from the "head" of the reader queue
 *      			(_b[READER].bufs is base address of queue).  This thread is
 *                  only launched for yuv 4:2:2 data using -cc sw|cm.  For -cc cm,
 *                  the openGL color matrix is used and each field is expanded to 
 *                  YUVA, 8 bits per component.  If software color space conversion 
 *                  is used (-cc sw) then each yuv 422 field is converted into RGBA. 
 *
 *	         		In either case, the modified data is placed at the "tail"
 *	           		of the converter queue (adress _b[CONVERTER].bufs).  If the -inter 
 *      			flag is specified, then two fields are expanded and interlaced 
 *      			within the same buffer.  From here data flows to:
 *
 *                                      |
 *                                      |       
 *                                    \ | /     
 *                                     \|/      
 *                                      V       
 *
 *	3. PutField() - This thread grabs data from the "head" of output queue it is
 *      			attached to (the reader queue if doing RGBA, RGBP or YUVP, OR the converter
 *      			queue if doing YUV 4:2:2 with -cc sw|cm).  The data is loaded into texture 
 *                  memory and a texture mapped quad is drawn. 
 */

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <windows.h>
#include <time.h>
#include <GL\gl.h>
#include "..\common\handy.h"
#include "..\common\fileutils.h"
#include "..\common\filepacking.h"

#define MAX_IOV 16 
#define READER 0
#define CONVERTER 1
#define NUM_BUFFS 2

typedef enum {_CC_NONE,          //No color conversion (rgb or rgba data) 
              _CC_SW_422,        //Color convert YUV 422->RGBA via software 
              _CC_CM_422,        //Color convert YUV 422->RGBA via color matrix 
              _CC_CM_444,        //Color convert YUV 444->RGB  via color matrix 
              _CC_DIR_422,       //Pass YUV 422 data directly to gfx (extension needed).
              } ColorConvert_t;

ColorConvert_t _ccstat;

/* There are two "buffers" used by the program.  One by the reader thread
 * to queue fields read from disk.  The other is used by the converter
 * thread when expanding out 4:2:2 data prior to RGB conversion.
 */

typedef struct buffer {
    char *bufs;         /* Each thread deposits fields or frames in this buffer */
    int   head;         /* Where consumer thread retrieves fields from. */
    int   tail;         /* Where producer thread places  fields it has processed*/
    int   num_bufs;     /* Number of fields/frames in buffer.	*/ 
    int   buf_size;     /* Size of each element in buffer.	*/ 
} buffer_t;

buffer_t _b[2];

/* _b[READER] = reader buffer, loaded by ReadIt() and drained by
 *		        either Expander() (for _CC_SW_422 and _CC_CM_422) 
 *              or by PutField() for all other states.
 *
 * _b[CONVERTER] = converter buffer (_CC_SW_422, _CC_CM_422). Loaded by
 *		           Expander() and drained by PutField(). 
 */

/* All global variables in this file are preceded by an "_".  */

/* Matrix stored in column major order.
 * Data is stored CbYCr where Cb = U, Cr = V.
 */
GLfloat _colorMat[16] = { 1.596027,   -0.812968,    0.000000,     0.0,
                          1.164383,    1.164384,    1.164384,     0.0,
                          0.000000,   -0.391762,    2.017232,     0.0,
                         -0.870787,    0.529591,   -1.081390,     0.0 };

/* Misc. declarations.	*/
int _winx = 0;		/* x dimension of gfx window.		*/
int _winy = 0;		/* y dimension of gfx window.		*/
int _CurField = 0;	/* Current field being processed.	*/
int _dbuffer = 1;	/* Double buffering flag.		*/
int _iovcnt = 12;	/* Number of reads per readv call.	*/
int _IamPlaying;    /* Am I playing or am I stopped?	*/
fileHeader_t _f;

/* Gfx declarations  */
HDC        _deviceContext;
HGLRC      _renderContext;
int 	   _interlace = 0;
int 	   _gfx_vSize;

/* Texture declarations. */
int _tex_width;
int _tex_height;
int _tex_format;
int _texA;

/* External subroutines: look in ..\common\common.dll */
void SetupConsole();
void createTexture(unsigned int *texname, int tex_width, int tex_height, int tex_format, int mip);
int  BuildGfxLists(int start, int u, int v, int tw, int th, int rhs);
void init_422_tables();

/* Local subroutines    */
void Quit(HWND);
void initGL(HWND);
void doit();
void exposeCB(HWND, PAINTSTRUCT *);
DWORD WINAPI PutField(LPVOID);
DWORD WINAPI ReadIt(LPVOID);
DWORD WINAPI Expander(LPVOID);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
ColorConvert_t HowToColorConvert(int, char *, int *);


/* Main entry point from windows */

int APIENTRY WinMain(HINSTANCE _instance, HINSTANCE _prevInst, LPSTR _cmdLine,
                     int _cmdShow)
{
    int i,n;
    int packing;
    char *Fname;
    char *cc_meth = NULL;
    char *className = "OpenGL";
    char *windowName = "tplay_u";
    MSG msg;
    WNDCLASSEX wndClass;
    HWND wnd;

    /* Set up standard IO for printing to a console window.*/
    SetupConsole();

    for (i=1; i < __argc; i++) {
    
        if (strcmp (__argv[i], "-single") == 0) {
            _dbuffer = 0;
        }
        else if (strcmp (__argv[i], "-inter") == 0) {
            _interlace = 1;
        }
        else if (strcmp (__argv[i], "-cc") == 0) {
            ++i; 
            cc_meth = __argv[i];
        }
        else if (strcmp (__argv[i], "-gfx") == 0) {
            ++i;
            _winx = atoi(__argv[i]);
            ++i;
            _winy = atoi(__argv[i]);
        }
        else if (strcmp (__argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(__argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
        else {
            Fname = __argv[i];
        }
    }

    // Initialize file
    fileInit(Fname, &_f, TRUE);
   
    // Determine color converint state.
    _ccstat = HowToColorConvert(_f.packing, cc_meth, &_tex_format);
    
    if (_interlace) {
        ERR( ((_ccstat != _CC_CM_422) && (_ccstat != _CC_SW_422)), "Use -inter for sw or cm 4:2:2");
    }

    // Cheat a liitle here, it makes the reading a lot simpler.
    if (_f.Nfields < _iovcnt) {
        _iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }
	
    if (_winx == 0) {
        _winx = _f.uSize; // Window size, texture will scale to this size.
        _winy = _f.vSize;
    }
    _gfx_vSize  = _f.vSize;

    // Initialize state for the reader buffers. 
    _b[READER].num_bufs = 3*_iovcnt; 
    _b[READER].buf_size = _f.readSize; 
    _b[READER].bufs =  (char *) VirtualAlloc(NULL, _b[READER].num_bufs * _b[READER].buf_size,
                                             MEM_COMMIT,PAGE_READWRITE);

    ERR ((_b[READER].bufs == NULL), "Failed to allocate reader buffers");


    // Initialize the state for the converter buffers - required for _CC_SW_422 or _CC_CM_422. 
    if ((_ccstat == _CC_SW_422) || (_ccstat == _CC_CM_422)) {  
        _b[CONVERTER].num_bufs = _iovcnt;
        _b[CONVERTER].buf_size = 2*_f.readSize;  //Converting YUV 4:2:2 to GL_RGBA.	

        if (_interlace) {
	        _b[CONVERTER].num_bufs /= 2;  // Need half then number of buffers if interlacing.
            _b[CONVERTER].buf_size *= 2;  // But each buffer is now twice as big.
            _gfx_vSize *=  2;             // The data dimensions are now twice the original size.
        }

        _b[CONVERTER].bufs = (char *) VirtualAlloc(NULL, 
                                                   _b[CONVERTER].num_bufs * _b[CONVERTER].buf_size,
                                                   MEM_COMMIT,PAGE_READWRITE);

        ERR ((_b[CONVERTER].bufs == NULL), "Failed to allocate converter buffers"); 
    }

    fprintf(stderr, "Playing  %d fields of %dx%d using a read vector of length %d\n", 
                     _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Padded read size (bytes) = %d \n", _f.readSize);
    fprintf(stderr, "Gfx window size = %dx%d\n", _winx, _winy);


    /* Do that nasty Windows stuff then head off to the message main loop.	*/
    wndClass.cbSize        = sizeof (WNDCLASSEX) ;
    wndClass.style         = CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc   = WndProc;
    wndClass.cbClsExtra    = 0 ;
    wndClass.cbWndExtra    = 0 ;
    wndClass.hInstance     = _instance ;
    wndClass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
    wndClass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
    wndClass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wndClass.lpszMenuName  = NULL ;
    wndClass.lpszClassName = className ;
    wndClass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

    /* Create window */
    RegisterClassEx (&wndClass) ;
    wnd = CreateWindow( className, windowName,
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                            100,                            // initial x position
                            100,                            // initial y position
                            _winx,                          // width
                            _winy,                          // Height
                            NULL,                           // parent window handle
                            (HMENU) NULL,                   // window menu handle
                            _instance,                      // program instance handle
                            NULL) ;

    ShowWindow(wnd, _cmdShow);
    UpdateWindow(wnd);

    /* Now loop here and process messages to the window.  */
    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg) ;
        DispatchMessage (&msg) ;
    }

    return msg.wParam ;

}



LRESULT CALLBACK WndProc(HWND _wnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
    /* Routine to process window events.        */
    PAINTSTRUCT ps;

    // Process the event.
    switch(_msg) {
        case WM_CREATE:
            initGL(_wnd);
            return 0;

        case WM_LBUTTONUP:
        case WM_MBUTTONUP:
        case WM_RBUTTONUP:
            doit();
            return 0;

        case WM_SYSCOMMAND:
            if(_wParam == SC_CLOSE) {
                fprintf(stderr, "Quitting ...\n");
                Quit(_wnd);
                return 0;
            }
            else break;

        case WM_PAINT:
            exposeCB(_wnd, &ps);
            return 0;
    }

    return DefWindowProc (_wnd, _msg, _wParam, _lParam) ;
}



void initGL(HWND wnd)
{
    /* This is the initialize call back routine for our widget.   This
     * is called automatically from WndProc callback.
     */

    int val;
    PIXELFORMATDESCRIPTOR   pfd;
    int     pixelFormat;

    pfd.nSize =                 sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion =              1;
    pfd.dwFlags =               PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_TYPE_RGBA;
    pfd.cColorBits =            24;
    pfd.cRedBits =              0;
    pfd.cRedShift =             0;
    pfd.cGreenBits =            0;
    pfd.cGreenShift =           0;
    pfd.cBlueBits =             0;
    pfd.cBlueShift =            0;
    pfd.cAlphaBits =            8;
    pfd.cAlphaShift =           0;
    pfd.cAccumBits =            0;
    pfd.cAccumRedBits =         0;
    pfd.cAccumGreenBits =       0;
    pfd.cAccumBlueBits =        0;
    pfd.cAccumAlphaBits =       0;
    pfd.cDepthBits =            8;
    pfd.cStencilBits =          1;
    pfd.cAuxBuffers =           0;
    pfd.iLayerType =            PFD_MAIN_PLANE;
    pfd.bReserved =             0;
    pfd.dwLayerMask =           0;
    pfd.dwVisibleMask =         0;
    pfd.dwDamageMask =          0;

    if (_dbuffer) {
        pfd.dwFlags |= PFD_DOUBLEBUFFER;
    }

    // Get the device context for the window.
    NC (_deviceContext = GetDC(wnd));

    // Match a supported pixel format to the specified pixel format.
    pixelFormat = ChoosePixelFormat(_deviceContext, &pfd);

    // Set the pixel format of the device context.
    ZC (SetPixelFormat(_deviceContext, pixelFormat, &pfd));

    // Create OpenGL rendering context and make it current.
    NC (_renderContext = wglCreateContext(_deviceContext));

    ZC (wglMakeCurrent(_deviceContext, _renderContext));

    _tex_width = 1024;
    _tex_height = (_f.vSize <= 256) ? 256 : 512;
    if (_interlace) _tex_height *= 2;	/* If interlacing, texture is twice as big */ 

    fprintf(stderr, "Texture width = %d Texture height = %d\n", _tex_width, _tex_height);

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    createTexture(&_texA, _tex_width, _tex_height, _tex_format, 0);

    /* init viewport */
    glViewport(0, 0, _winx, _winy);

    /* set up transformation matrix */
    if (_ccstat == _CC_SW_422) {
        fprintf(stderr, "***** Doing 422 Software Color Space Conversion *******\n");
        init_422_tables();
    }
    else if ((_ccstat == _CC_CM_422) || (_ccstat == _CC_CM_444)) {
        fprintf(stderr, "***** Setting OpenGL Color Matrix Color *******\n");
        glMatrixMode(GL_COLOR);
        glLoadMatrixf(_colorMat);
    }
    else if (_ccstat == _CC_DIR_422) {
        fprintf(stderr, "***** Setting OpenGL Pixel Unpacking for YUV 422 *******\n");
        glPixelStorei(GL_UNPACK_SUBSAMPLE_RATE_SGIX, GL_PIXEL_SUBSAMPLE_4242_SGIX);
    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glColor4f(1.0, 1.0, 1.0, 1.0);
    glDisable(GL_DITHER);

    (void)BuildGfxLists(1, _f.uSize, _gfx_vSize, _tex_width, _tex_height,1);

    /* enable texturing */
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,_texA);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }
}



void exposeCB(HWND wnd, PAINTSTRUCT *ps)
{
    BeginPaint(wnd, ps);
    if (_IamPlaying == 0) {

        ZC (wglMakeCurrent(_deviceContext, _renderContext));

        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) SwapBuffers(_deviceContext);

        glClear(GL_COLOR_BUFFER_BIT);
        glFlush();
    }
    EndPaint(wnd, ps);
}


void doit()
{
    /* This routine is called each time the left mouse button is clicked.
     * If things are stopped then the three play threads are startedd. 
     * If things are playing then the the threads are told to exit by 
     * setting _IamPlaying to 0.
     */

    int i;
    int expanding;
    int allDone;
    HANDLE readT, putT, expT;
    DWORD readID, putID, expID;

    expanding  = (_ccstat == _CC_SW_422) || (_ccstat == _CC_CM_422);

    if (_IamPlaying == 0) {  /* Go and play */
        _IamPlaying = 1;

        fprintf(stderr, "Starting play\n");
        for (i=0; i < NUM_BUFFS; i++) {
            _b[i].head = 0; 
            _b[i].tail = 0; 
        }

        /* Surrender the graphics context.      */
        ZC (wglMakeCurrent(_deviceContext, NULL));

        /* Start the read thread.   */
        NC(readT = CreateThread (NULL, 0, ReadIt, (LPVOID *)&(_b[READER]), 0, &readID));
        i = READER;

        if (expanding) { 
            fprintf(stderr, "**** Launching converter thread\n");
            i = CONVERTER; 
            NC(expT = CreateThread (NULL, 0, Expander, (LPVOID *)&(_b[0]), 0, &expID));
        }

        // Depending on the value of "i", PutField talks to either the converter or the reader
        NC(putT = CreateThread (NULL, 0, PutField, (LPVOID *)&(_b[i]), 0, &putID));
    }

    else { 
        allDone = (expanding) ? -3 : -2;
        _IamPlaying = 0;

        while (_IamPlaying != allDone) Sleep(10);  /* Wait for two or three threads */
        _IamPlaying = 0;

        ZC (wglMakeCurrent(_deviceContext, _renderContext));  /* Grab context */

        /* Make sure _CurField is pointing at an even field	*/
        if ((_CurField % 2) != 0) {
            _CurField++;
            if (_CurField >= _f.Nfields) _CurField = 0; 
        }
        fprintf(stderr, "Play stopped\n");
    }
}


DWORD WINAPI PutField(LPVOID param)
{
    buffer_t *b;
    int tmp;
    int x,y;
    int count = 0;
    char *dataPtr;
    double duration;
    static clock_t start, finish;

    /* This thread exits second and sets _IamPlaying to -2.  */

    /* This thread takes data from the head of the buffer it is
     * attached to.  In other words, input parameter "*b" equals either
     * 	1. &_b[CONVERTER] for yuv 422_8 data on disk which is being converted 
     *     using either _CC_SW_422 or _CC_CM_422 by the thread Expander(). 
     *		OR
     *  2. &_b[READER] for  all other _CC* states.
     * 
     * Data is loaded into texture memory and a texture mapped quad is drawn.
     */

    b = (buffer_t *)param; /* Input prameter is on of buffers described above */

    /* Set thread to run on cpu 1 ????? */ 
    
    ZC (wglMakeCurrent(_deviceContext, _renderContext));  /* Grab context */

    x = _tex_width  - _f.uSize;
    y = _tex_height - _gfx_vSize;

    while (_IamPlaying != -1) {

        if (b->head == b->tail) {
            //fprintf(stderr, "PutField():  input queue is empty %d\n", count); 
            Sleep(10);
            continue;
        }

        /* Grab the next valid buffer from the head of the buffer*/ 
        dataPtr = b->bufs + b->head * b->buf_size;

        /* Load texture memory and draw the quad.	*/
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, _f.uSize, _gfx_vSize, _tex_format, 
                        GL_UNSIGNED_BYTE, dataPtr);
        glCallList(1);
        
        if (count == 0) {
            start = clock();
        }
        count++;

        if (_dbuffer) {
            if (!_interlace) {
                /* Draw black along top and bottom to hide half lines.	*/
                glDisable(GL_TEXTURE_2D);
                glColor4f(0.0,0.0,0.0,0.0);
                glRectf(-1.0, 1.0, 1.0, 0.98);
                glRectf(-1.0, -1.0, 1.0, -0.98);
                glEnable(GL_TEXTURE_2D);
                glColor4f(1.0, 1.0, 1.0, 1.0);
            }
            SwapBuffers(_deviceContext);
        }

        /* Advance head pointer to next buffer location.	*/
        tmp = b->head + 1;
        if (tmp == b->num_bufs) tmp = 0;
        b->head = tmp;

    }  /* End of while (IamPlaying) */ 

    finish = clock();
    duration = (double)(finish - start)/CLOCKS_PER_SEC;
    fprintf(stderr, "Draw time = %f sec. fps = %f\n", duration, ((double)(count))/duration);

    /* Surrender the graphics context.      */
    ZC (wglMakeCurrent(_deviceContext, NULL));

    _IamPlaying--;		            /* Tell doit() all is done */	
    ExitThread (-1);
}



DWORD WINAPI ReadIt(LPVOID param)
{
    /* This thread exits first and sets _IamPlaying to -1.  */

    /* This thread reads image data from disk and places it
     * at the tail of the buffer pointed to by the "*b" input
     * parameter (which is always &_b[READER]).  This data will then 
     * be grabbed  by:
     *	   1. The converter thread (see Expander()) for _CC_SW_422 or _CC_CM_422. 
     *			OR
     *	   2. The display thread (see PutField()) for all other _CC* states.
     */

    buffer_t *b;
    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;
    LONG hi_off, lo_off;
    DWORDLONG offset;

    b = (buffer_t *)param;  /* Grab input buffer */

    offset = _f.headerSize + (DWORDLONG)_CurField * _f.readSize; 
    hi_off = (LONG)(offset>>32);
    lo_off = (LONG)(offset & 0x00000000ffffffff);
    SetFilePointer(_f.fd, lo_off, &hi_off, FILE_BEGIN);

    while (_IamPlaying) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Now make sure there is room in reader buffer to store 
         * "fieldsToRead" fields (which the consumer PutField() will later
         * display. 
         */
        do {
            full = 0;
            /* Check for the first full condition.  If the incremented
             * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
            tmphead = b->head;
            tmptail = b->tail + fieldsToRead;
            if (tmptail == b->num_bufs) tmptail = 0;
            full = (tmptail == tmphead); 

            /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
            full |= ((tmphead > b->tail) && (tmphead < b->tail + fieldsToRead));

            if (full) {
                //fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail);
                Sleep(10);
            }
        } while (full);

        /* Read the fields.		*/
        tmptail = b->tail;
        ZC ( ReadFile(_f.fd, b->bufs + tmptail*b->buf_size, fieldsToRead*b->buf_size, &i, NULL));
        tmptail += fieldsToRead;

        /* Inc. tail, indicating more vaild data is in buffer.	*/
        if (tmptail == b->num_bufs) tmptail = 0;
        b->tail = tmptail;
	
        /* Check for file wrap.		*/
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
            _CurField = 0;
    	    SetFilePointer(_f.fd, _f.headerSize, NULL, FILE_BEGIN);
            fprintf(stderr, "Done pass through file\n");
        }

   }  /* end while(IamPLaying) */

   _IamPlaying--;

   ExitThread (-2);
}


DWORD WINAPI Expander(LPVOID param)
{
    /* This thread exits last (third) and sets _IamPlaying to -3 */

    /* This routine takes fields from the head of reader buffer and
     * does one of the following (this thread only used for _CC_SW_422
     * and _CC_CM_422). 
     *     1. Expands the 4:2:2 field into yuva 4:4:4:4 if color matrix is
     *        enabled (_CC_CM_422).  
     *				OR
     *     2. For _CC_SW_422, the 4:2:2 field is converted to
     *	      RGBA using the cpu.
     *
     * The new result is placed at the tail of the converter buffer.  
     * The PutField thread takes from the head of converter buffer and draws the 
     * final image to the screen. Note:
     *
     *			b[READER] = reader buffer.
     *  		b[CONVERTER] converter buffer. 
     *
     * If _interlace is TRUE then field expands two yuv fields and interlaces
     * them in the buffer within the converter buffer.  This program uses
     * the variable "field" to determine if it is working on the first (value of 0)
     * or second (value of 1) yuv input field for a particular frame.  If _interlace
     * is FALSE then "field" remains 0 in this code.
     */

    int tmphead;
    int tmptail;
    int field = 0;
    char *yuvPtr;
    char *rgbPtr;
    buffer_t *b;

    b = (buffer_t *)param;
    
    while (_IamPlaying != -2) {

        if (b[READER].head == b[READER].tail) {
            // fprintf(stderr, "input queue is empty.\n"); 
            Sleep(10);
            continue;
        }

        /* If not interlacing OR interlacing and it is the first field of
         * of the pair THEN go and grab the next free converter buffer
         * from the tail of that queue.
         */
        if (field == 0) {
            tmptail = b[CONVERTER].tail + 1;
            if (tmptail == b[CONVERTER].num_bufs) tmptail = 0; 
            if (tmptail == b[CONVERTER].head) {       
                // fprintf(stderr, "No room in converter queue\n");
                Sleep(10);
                continue;
            }
        }

        /* Grab the next valid buffer from the head of the "READER list
         * (yuvPtr) and add to end of converted list (rgbPtr).
         */
        yuvPtr  = b[READER].bufs + b[READER].head * b[READER].buf_size;
        rgbPtr  = b[CONVERTER].bufs + b[CONVERTER].tail * b[CONVERTER].buf_size;

        /* If interlacing and it is the first of the pair then start the
         * expansion at the correct scan line.
         */
        if (_interlace && (field == 0) ) {
            rgbPtr += _f.uSize * 4;
        }

        if (_ccstat == _CC_SW_422) {
            fast422ToRGB(_f.uSize, _f.vSize, yuvPtr, rgbPtr, _interlace);
            /* fast422ToRGB_Filter(_f.uSize, _f.vSize, yuvPtr, rgbPtr, _interlace); */
        }
        else {
            expandyuv(_f.uSize, _f.vSize, yuvPtr, rgbPtr, _interlace);
        }

        /* Advance head of reader buffer.	*/
        tmphead = b[READER].head + 1;
        if (tmphead == b[READER].num_bufs) tmphead = 0;
        b[READER].head = tmphead;

        if (_interlace == 0) {
            b[CONVERTER].tail = tmptail;
        }
        else {
            /* If interlacing and you did the second field, then it is time
             * to send the buffer for drawing.
             */ 
            if (field == 1) b[CONVERTER].tail = tmptail;
            field = (field) ? 0 : 1;
        }


    }  /* End of while (IamPlaying) */ 

    _IamPlaying--;
    ExitThread(-3);
}

void Quit(HWND wnd)
{
    int i;

    /* Stop play */
    if (_IamPlaying) doit();

    wglMakeCurrent(_deviceContext, _renderContext);
    wglDeleteContext(_renderContext);

    ReleaseDC(wnd, _deviceContext);
    CloseHandle(_f.fd);
    PostQuitMessage(0);
}



ColorConvert_t HowToColorConvert(int packing, char *cc_meth, int *pixelFormat)
{

    ColorConvert_t cstate;

    /* This program looks at the file packing and the command line string
     * passed in through "cc_meth" and returns the color convertng state for
     * the rest of the program to use.  It also determines the pixel format 
     * that will be loaded into texture memory.
     */

    switch (packing) {

    case VL_PACKING_BGR_8_P :
        *pixelFormat = GL_RGB;
        cstate = _CC_NONE;
        break;

    case VL_PACKING_ABGR_8 :
        *pixelFormat = GL_RGBA;
        cstate = _CC_NONE;
        break;

    case VL_PACKING_YVYU_422_8 : 
        if ((cc_meth == NULL)  || (strcmp(cc_meth,"cm") == 0)) {
            cstate = _CC_CM_422;
            *pixelFormat = GL_RGBA;           //422 expanded to YUVA & converted by GL color matrix.
        }
        else if (strcmp(cc_meth,"sw") == 0) {
            cstate = _CC_SW_422;
            *pixelFormat = GL_RGBA;           //422 --> RGBA by fast422ToRGB() software routine.
        }
        else if (strcmp(cc_meth,"dir") == 0) {
            cstate = _CC_DIR_422;
            *pixelFormat = GL_YCRCB_SGIX; //422 passed directly to openGL. 
        }
        else {
            ERR(1, "Invalid -cc option");
        }
        
        break;

    case VL_PACKING_UYV_8_P :
        if ((cc_meth == NULL)  || (strcmp(cc_meth,"cm") == 0)) {
            cstate = _CC_CM_444;
            *pixelFormat = GL_RGB;            //444 converted to RGB by GL color matrix.
        }
        else if (strcmp(cc_meth,"sw") == 0) {
            ERR(1, "Software conversion for 422 only");
        }
        else if (strcmp(cc_meth,"dir") == 0) {
            ERR(1, "Direct support for 422 only");
        }
        else {
            ERR(1, "Invalid -cc option");
        }
        break;

    default: 
        ERR(1,"Bad packing");
        break;
  
    }
    return cstate;
}
