/*
 * Copyright (c) 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vgrab_u = Grabs uncompressed video to disk. 
 * 
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of fields written per WriteFile call.
 *						          Default is 12, 8 and 16 are also good values.
 *     <-n fieldsToCapture> 	- Number of fields to capture, default = 200.
 *								  This arg should be an even number.
 *     <-rgba OR				- Capture RGBA fields to disk (40 MB/sec!!)
 *		-422_10>  	            - Capture YUV 4:2:2 10 bit to disk (40 MB/sec.). 
 *								- Default is YUV 4:2:2 8 bit.
 *
 *     <-frames>		 		- Capture frames instead of fields. 
 *     <-sv>		 		    - Use svideo input, default is composite. 
 *
 *     fileName                 - File for results. 
 *
 * This is a real simple program to capture field/framess of uncompresed video
 * to disk in real time.  Unbuffered IO (with WriteFile call) is used to move
 * the data to disk.  Fields are rounded to 16k bytes, which satisfies unbuffered
 * IO constraints for both NT and IRIX.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() is called to
 *        perform the real work.
 *		- doit() sits in a tight loop grabbing fields from live video (via
 *	      the rotuine checkReceiveQueue()). 
 *		- Once "_iovcnt" fields have been accumulated, they are written to disk
 *	  	  using WriteFile().
 *		- After capturing the desired number of fields, the program exits.
 */


#include <stdio.h>
#include <stdlib.h>
#include <sgidmedia.h>
#include "..\common\handy.h"
#include "..\common\dmhandy.h"
#include "..\common\fileutils.h"
#include "..\common\filepacking.h"

// When we send buffers to dmPath via dmPathSendBuffers, use
// the following indices for the image buffer, UST, and MSC.
#define _VC_BUFFER_INDEX    0
#define _VC_UST_INDEX       1
#define _VC_MSC_INDEX       2
#define _VC_BUFFER_PV_END   3

#define MAX_IOV 16

// All global variables in this file are preceded by an "_".
char         *_readvBufs;		// Pointer to buffers holding video samples.

int  _iovcnt = 12;		// # of fields per write.	
int	 _capType = DM_INTERLEAVE_MODE_SINGLE_FIELD;	
int  _mscStep;

fileHeader_t *_f;
HDMPATH		  _hVideoPath;

// Procudure declarations
void doit();
void initVideo(dmJackTypeEnum );
fileHeader_t *fileCreate(char *fname, int unbuffered); 

main(int argc, char *argv[])
{
    int i;
    int  imageCount = 200;
    int packing = DM_FORMAT_CbYCrY_8_601;
    dmJackTypeEnum jack =  DM_JACK_TYPE_COMPOSITE;
    char * fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            imageCount = atoi(argv[i]);
	    }
		else if (strcmp (argv[i], "-iov") == 0) {
	    	++i;
	    	_iovcnt = atoi(argv[i]);
	    	if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
		}
    	else if (strcmp (argv[i], "-rgba") == 0) {
	    	packing = DM_FORMAT_RGBA_8_FULL;
		}
    	else if (strcmp (argv[i], "-422_10") == 0) {
	   		packing = DM_FORMAT_CbYCrY_10L_601; 
		}
		else if (strcmp (argv[i], "-frames") == 0) {
	    	_capType = DM_INTERLEAVE_MODE_INTERLEAVED;
		}
		else if (strcmp (argv[i], "-sv") == 0) {
            jack =  DM_JACK_TYPE_SVIDEO;
        }
		else {
	    	fname = argv[i];
		}
    }

    _f = fileCreate(fname, TRUE);
    _f->packing = packing;
    _f->compressed = FALSE;
    _f->pad = _f->headerSize;
    _f->Nfields = imageCount;

    if (_iovcnt > imageCount) _iovcnt = imageCount;

    fprintf(stderr, "Capturing %d fields,  writev vector length = %d\n", 
            imageCount,_iovcnt);


    initVideo(jack);
    doit();
}


void doit()
{
    /* The routine which does all the work capturing and writing 
     * fields to disk.
     */
    DMint64 oldmsc = -1;
    DMint64 newmsc;
    DMint64 ust;
    int i, toGet;
    int fieldsCaptured = 0;
    char *dataPtr, *writePtr;
    DWORD bytesWritten;
    DMpv p[10];

    // Start video samples flowing along path. 
    DC (dmPathBeginTransfer(_hVideoPath) );

    /* Now loop and capture the fields. */
    while (fieldsCaptured < _f->Nfields) {

		toGet = ((fieldsCaptured + _iovcnt) <= _f->Nfields) ?
                         _iovcnt : (_f->Nfields - fieldsCaptured);

        for (i= 0; i < toGet; i++) {
            do {    /* Get pointer to next valid field */ 
				checkReceiveQueue(&dataPtr, &ust, &newmsc);
            } while (dataPtr == NULL);

			if (i== 0) writePtr = dataPtr;

            if ((oldmsc != -1) && (newmsc != oldmsc + _mscStep)) {
                fprintf(stderr, "MSC drop old = %lld, ", oldmsc);
                fprintf(stderr, "new = %lld\n", newmsc);
            }
            oldmsc = newmsc;
		}

        /* Remember, the buffers are all allocated from a contiguous chunk of memory
         * and sent to the system in order - so the application gets them back in order.
         * Since the code writes _iovcnt buffers at a time, AND since the total buffer pool
         * is a multiple of _iovcnt, the code can write _iovcnt buffers to disk as one
         * contiguous write.  The only time _iovcnt is not written is the very last pass,
         * in which case _f->Nfields % _iovcnt are written.
         */
	    ZC (WriteFile(_f->fd, writePtr, (DWORD)toGet*_f->readSize, &bytesWritten, NULL));

		// Indicate that the buffers are free for re-use.
        for (i=0; i < toGet; i++) {
        	p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
        	p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)(writePtr + i*_f->readSize);
        	p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _f->readSize; 
        	p[_VC_UST_INDEX].param = DM_VIDEO_UST;
        	p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
        	p[_VC_BUFFER_PV_END].param = DM_END;

        	DC (dmPathSendBuffers(_hVideoPath, p));
		}
		fieldsCaptured += toGet;
       
        if (fieldsCaptured % 600 == 0) {
	    	fprintf(stderr, "...Captured %d fields\n", fieldsCaptured);
		}
    }

	DC (dmPathEndTransfer(_hVideoPath));

	DC (dmPathClose(_hVideoPath));

    /* Write out the correct header info.	*/
    _f->packing = DMFormatToFilePacking(_f->packing);
    writeHeader(_f);
	CloseHandle(_f->fd);

    fprintf(stderr, "Done writing %d fields to file\n", fieldsCaptured);
}




void initVideo(dmJackTypeEnum jacktype)
{
    int i, numbufs;
    int bytes_per_pixel;
    HDMJACK hJack;
    DMpv *sysCaps, *jackCaps, *pathCaps;
    DMint64 *pJacks, *pPaths;
    DMint32 nJacks, nPaths;
    DMpv p[10];

    if (_f->packing == DM_FORMAT_CbYCrY_8_601) {
        bytes_per_pixel = 2;
    }
    else {
        bytes_per_pixel = 4;
    }

    DC (dmSystemGetCapabilities(DM_SYSTEM_LOCALHOST, &sysCaps));
    pJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].value.pInt64;
    nJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].length;

    // Find a jack of the given type, i.e. composite, digial, etc.).
    for (i = 0; i < nJacks; i++) {
        DC (dmJackGetCapabilities(pJacks[i], &jackCaps));

        if (jackCaps[DM_JACK_DIRECTION_INDEX].value.int32 == DM_DIRECTION_IN &&
            jackCaps[DM_JACK_TYPE_INDEX].value.int32 == jacktype){
            break;
        }
    }

    ERR ((i == nJacks), "initVideo: can't find a jack");

    // Make sure there is a signal present on the desired jack.
    DC (dmJackOpen(pJacks[i], NULL, &hJack));
    i = 0;
    p[i].param = DM_JACK_SIGNAL_PRESENT_INT32; i++;
    p[i].param = DM_END;                       i++;
    DC (dmJackGetControls(hJack, p));
    ERR ((p[0].value.int32 == DM_TIMING_NONE), "No signal present for desired jack");
    ERR ((p[0].value.int32 == DM_TIMING_UNKNOWN), "Unknown signal present for desired jack");

    // For the given jack, find the dev to mem path.
    pPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].value.pInt64;
    nPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].length;
    for (i = 0; i < nPaths; i++) {
        DC (dmPathGetCapabilities(pPaths[i], &pathCaps));
        if (pathCaps[DM_PATH_TYPE_INDEX].value.int32 == DM_PATH_TYPE_DEV_TO_MEM) 
            break;
    }

    ERR ((i == nPaths), "initVideo: can't find a path");

    DC (dmPathOpen(pPaths[i], NULL, &_hVideoPath));

    //Set the image format and interleave mode on the path.
    i = 0;
    dmPvSetInt32Param( p[i], DM_IMAGE_FORMAT_INT32, _f->packing); i++;
    dmPvSetInt32Param( p[i], DM_IMAGE_INTERLEAVE_MODE_INT32, _capType); i++;
    dmPvSetInt32Param( p[i], DM_END, 0 ); i++;
    DC (dmPathSendControls(_hVideoPath, p));

    //Now read back the settings needed for the file header.
    i = 0;
    p[i].param = DM_VIDEO_WIDTH_INT32;     i++;
    p[i].param = DM_VIDEO_HEIGHT_F1_INT32; i++;
    p[i].param = DM_VIDEO_HEIGHT_F2_INT32; i++;
    p[i].param = DM_END;                   i++;
    DC (dmPathQueryControls(_hVideoPath, p));

    _f->uSize = p[0].value.int32;	//image width;
    _f->vSize = p[1].value.int32;	//image height;

    if (_capType == DM_INTERLEAVE_MODE_INTERLEAVED) {
        _f->vSize += p[2].value.int32; //Frames
        _mscStep = 2;
    }
    else {
        _mscStep = 1;
    }

    // Round io size up to the length of _f->headerSize.
    _f->readSize = bytes_per_pixel * _f->vSize * _f->uSize;
    i = _f->readSize/_f->headerSize;
    if ((_f->readSize % _f->headerSize) != 0) ++i;
    _f->readSize = _f->headerSize*i;

    fprintf(stderr, "usize = %d vsize = %d readsize = %d, msc Step = %d\n", 
                     _f->uSize, _f->vSize, _f->readSize, _mscStep);

    /* Allocate the buffers used to capture video.
     * The buffers are allocated sequentially out of a contiguous chunk of
     * memory - this way they can be written back as a group later on.
     */

    numbufs = _iovcnt * 3;
    _readvBufs =  (char *) VirtualAlloc(NULL,  numbufs*_f->readSize, MEM_COMMIT,PAGE_READWRITE);
    ERR ((_readvBufs == NULL), "Failed to allocate reader buffers");


    // Send all available buffers to the video driver.  That way, when
    // the video transfer is started (see doit) there will be buffers 
    // ready to load.
    for (i = 0; i < numbufs; i++) {
        p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
        p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)(_readvBufs + i*_f->readSize);
        p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _f->readSize; 
        p[_VC_UST_INDEX].param = DM_VIDEO_UST;
        p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
        p[_VC_BUFFER_PV_END].param = DM_END;

        DC (dmPathSendBuffers(_hVideoPath, p));
    }
}



int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc)
{
    // Routine to check message queue attached to video hardware.  In
    // particular, this routine will return the address of a buffer 
    // just loaded with a field/frame from video input.   It also returns
    // the ust and msc of the buffer as well.
 
    DMint32 message;
    DMpv *reply;
    DMstatus ret;
    int rval = 0;
 
    *data = NULL;
    *ust = -1;
    *msc = -1;

    ret = dmPathReceiveMessage(_hVideoPath, &message, &reply);

    if (ret != DM_STATUS_NO_ERROR) {     //Did not get error free result.

		if (ret != DM_STATUS_RECEIVE_QUEUE_EMPTY) {  //Did not get empty queue
	    	fprintf(stderr, "dmPathReceiveMessage returned value %x\n", ret);
		}
		return -1;  //Either error or queue empty.
    }

    switch (message) {
    case DM_MESSAGE_TRANSFER_COMPLETE:
		*data = (unsigned char *)reply[_VC_BUFFER_INDEX].value.pByte;
		*ust = reply[_VC_UST_INDEX].value.int64;
		*msc = reply[_VC_MSC_INDEX].value.int64;
		break;
    case DM_MESSAGE_MISSED_FIELD:
        //fprintf(stderr, "DM_MESSAGE_MISSED_FIELD\n");
		break;
    case DM_MESSAGE_TRANSFER_ABORTED:
		*data = (unsigned char *)reply[0].value.pByte;
        fprintf(stderr, "DM_MESSAGE_TRANSFER_ABORTED\n");
		break;
	case DM_MESSAGE_CONTROLS_SUCCESSFUL:
        fprintf(stderr, "DM_MESSAGE_CONTROLS_SUCCESSFUL\n");
		break;
    default:
        fprintf(stderr, "dmPathReceiveMessage returned message %x\n", ret);
		break;
    }

    return rval;
}
