/*
 * Copyright (c) 1997-1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vidin2tex  - Grabs live video input, loads it into texture memory and 
 *              displays the result in a window. 
 *
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-single>		- Use single buffering 
 *     <-rgba>          - Convert video to RGBA on input (default is 4:2:2).
 *     <-sv>			- Use svideo input. Default is composite. 
 *     <-gfx winx winy> - Size of display window on gfx monitor.  The video 
 *                        fields are filtered to fit this window by texture 
 *                        mapping.
 *
 *
 * How to run the program:
 * =======================
 *     A white window appears.  Click in the window with any mouse button
 *     to display live video in the window.  Each time you click you start/stop
 *     the display of video in the window. 
 *
 * Here are the thread descriptions:
 * =================================
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() is called 
 *        when a mouse button is depressed.
 *		- If things are playing, doit() will cause play to stop.
 *      - If things are stopeed, doit() spawns the thread DrawVideo() to do the 
 *        real work.  The main thread then waits for the next button push. 
 *
 *    Draw Video:
 *      - Created by from the routine doit() in the main program thread.
 *      - This thread constantly grabs the next field from video input
 *        and loads it into texture memory.  
 *      - The field is then drawn as a texture mapped quad to fit the gfx 
 *        window (the size of which is specified with the -gfx flag).
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <windows.h>
#include <time.h>
#include <GL\gl.h>
#include <sgidmedia.h>
#include "..\common\handy.h"
#include "..\common\dmhandy.h"

// When we send buffers to dmPath via dmPathSendBuffers, use
// the following indices for the image buffer, UST, and MSC.
#define _VC_BUFFER_INDEX    0
#define _VC_UST_INDEX       1
#define _VC_MSC_INDEX       2
#define _VC_BUFFER_PV_END   3
#define _RVBUFS             4  //How many buffers for driving video


/* wgl defines */
HDC             _deviceContext;
HGLRC           _renderContext;

/* texture defines */
unsigned int _texA;                      /* texture names */
int _tex_width;                          /* texture size */
int _tex_height;                         /* texture size */
int _tex_format;

/* OpenGL defines */
int _dbuffer = 1;
int _uSize;
int _vSize;
int _winx = 0;
int _winy;
int _packing = DM_FORMAT_CbYCrY_8_601;

/* Play control variables */
int _IamPlaying = 0;

/* video buffer declarations  */
char     *_videoBufs;	// memory for holding buffers from video hardware. 
int       _videoSize;
HDMPATH   _hVideoPath;

/* External subroutines: look in ..\common\common.dll */
void SetupConsole();
void createTexture(unsigned int *texname, int tex_width, int tex_height, int tex_format, int mip);
int  BuildGfxLists(int start, int u, int v, int tw, int th, int rhs);

/* Local subroutines    */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void Quit(HWND);
void initGL(HWND);
void doit();
void exposeCB(HWND, PAINTSTRUCT *); 
DWORD WINAPI DrawVideo(LPVOID);
void initVideo(dmJackTypeEnum);
int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc);

/* Main entry point from windows */

int APIENTRY WinMain(HINSTANCE _instance, HINSTANCE _prevInst, LPSTR _cmdLine,
                     int _cmdShow)
{
    int i,n;
    MSG msg ;
    WNDCLASSEX wndClass;
    HWND wnd;
    dmJackTypeEnum jack =  DM_JACK_TYPE_COMPOSITE;
    char *className = "OpenGL";
    char *windowName = "vidin2tex";

    // Set up standard IO for printing to a console window.
    SetupConsole();

    // Parse arguments.
    for (i=1; i < __argc; i++) {
        if (strcmp (__argv[i], "-single") == 0) {
            _dbuffer = 0; 
        }
        else if (strcmp (__argv[i], "-rgba") == 0) {
	    	_packing = DM_FORMAT_RGBA_8_FULL;
        }
		else if (strcmp (__argv[i], "-sv") == 0) {
            jack =  DM_JACK_TYPE_SVIDEO;
        }
		else if (strcmp (__argv[i], "-gfx") == 0) {
            ++i;
            _winx = atoi(__argv[i]);
            ++i;
            _winy = atoi(__argv[i]);
        }
    }

    // Init video hardware.
    initVideo(jack);

    if (_winx == 0) {
        _winx = _uSize;
        _winy = _vSize;
    }
    fprintf(stderr, "Capturing %d x %d samples of %d bytes\n",_uSize, _vSize, _videoSize);
    fprintf(stderr, "Window size = %d x %d\n",_winx, _winy, _videoSize);

    wndClass.cbSize        = sizeof (WNDCLASSEX) ;
    wndClass.style         = CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc   = WndProc;
    wndClass.cbClsExtra    = 0 ;
    wndClass.cbWndExtra    = 0 ;
    wndClass.hInstance     = _instance ;
    wndClass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
    wndClass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
    wndClass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wndClass.lpszMenuName  = NULL ;
    wndClass.lpszClassName = className ;
    wndClass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

    /* Create window */
    RegisterClassEx (&wndClass) ;
    wnd = CreateWindow( className, windowName,
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                            100,                            // initial x position
                            100,                            // initial y position
                            _winx,                          // width
                            _winy + 20,                     // Height
                            NULL,                           // parent window handle
                            (HMENU) NULL,                   // window menu handle
                            _instance,                      // program instance handle
                            NULL) ;

    ShowWindow(wnd, _cmdShow);
    UpdateWindow(wnd);

    /* Now loop here and process messages to the window.  */
    while (GetMessage (&msg, NULL, 0, 0)) {
        TranslateMessage (&msg) ;
        DispatchMessage (&msg) ;
    }

    return msg.wParam ;
}




LRESULT CALLBACK WndProc(HWND _wnd, UINT _msg, WPARAM _wParam, LPARAM _lParam)
{
    // Routine to process window events. Enter here when buttons are depresed.
    PAINTSTRUCT ps;

    // Process the event.
    switch(_msg) {
        case WM_CREATE:
            initGL(_wnd);
            return 0;

        case WM_LBUTTONUP:
        case WM_MBUTTONUP:
        case WM_RBUTTONUP:
            doit();
            return 0;

        case WM_SYSCOMMAND:
            if(_wParam == SC_CLOSE) { 
                fprintf(stderr, "Quitting ...\n");
                Quit(_wnd); 
                return 0; 
            }
            else break;

        case WM_PAINT:
            exposeCB(_wnd, &ps);
            return 0;
    }

    return DefWindowProc (_wnd, _msg, _wParam, _lParam) ;
}




/* Initialize openGL rendering context */
void initGL(HWND wnd)
{
    int val;
    PIXELFORMATDESCRIPTOR   pfd;
    int     pixelFormat;


    memset (&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
    pfd.nSize =                 sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion =              1;
    pfd.dwFlags =               PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW;
    pfd.iPixelType =            PFD_TYPE_RGBA;
    pfd.cColorBits =            24;
    pfd.cDepthBits =            24;
    pfd.cStencilBits =          1;
    pfd.iLayerType =            PFD_MAIN_PLANE;

    if (_dbuffer) {
        pfd.dwFlags |= PFD_DOUBLEBUFFER;
    }

    // Get the device context for the window.
    NC (_deviceContext = GetDC(wnd));

    // Match a supported pixel format to the specified pixel format.
    pixelFormat = ChoosePixelFormat(_deviceContext, &pfd);

    // Set the pixel format of the device context.
    ZC (SetPixelFormat(_deviceContext, pixelFormat, &pfd)); 

    // Create OpenGL rendering context and make it current.
    NC (_renderContext = wglCreateContext(_deviceContext));

    ZC (wglMakeCurrent(_deviceContext, _renderContext));

    /* Set up texture type, width, and size */
	if (_packing == DM_FORMAT_RGBA_8_FULL)         _tex_format = GL_RGBA;
    else if (_packing ==  DM_FORMAT_CbYCrY_8_601)  _tex_format = GL_YCRCB_SGIX; 

    _tex_width = 1024; 
    _tex_height = (_vSize <= 256) ? 256 : 512; 

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
 
    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_format, 0);


    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    glEnable(GL_TEXTURE_2D);
    
    (void)BuildGfxLists(1, _uSize, _vSize, _tex_width, _tex_height,0);


    if (_tex_format ==  GL_YCRCB_SGIX) {
        fprintf(stderr, "***** Setting OpenGL Pixel Unpacking for YUV 422 *******\n");
        glPixelStorei(GL_UNPACK_SUBSAMPLE_RATE_SGIX, GL_PIXEL_SUBSAMPLE_4242_SGIX);
    }

    /* init viewport */
    glViewport(0,0,_winx,_winy);
    glDisable(GL_DITHER);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }

}




void exposeCB(HWND wnd, PAINTSTRUCT *ps) 
{
    BeginPaint(wnd, ps);
    EndPaint(wnd, ps);
}




void doit()
{
    int allDone, i;
    HANDLE playThread;
    DWORD  playThreadID;

    /* Routine to START/STOP play. 
     * This routine is entered directly when the left mouse button is clicked
     * or from loadit() when changing effects.
     */

    if (_IamPlaying) {
        _IamPlaying = 0;
        allDone = -1; 

        // Wait for both readers and the effect thread to exit.
        while (_IamPlaying != allDone) Sleep (10);
        _IamPlaying = 0;
 
        ZC (wglMakeCurrent(_deviceContext, _renderContext));

    }
    else {
        fprintf(stderr, "Starting play\n");
        _IamPlaying = 1;

        // Surrender the graphics context.
        ZC (wglMakeCurrent(_deviceContext, NULL));

        // Start the effect play back thread.
        NC(playThread = CreateThread (NULL, 0, DrawVideo, NULL, 0, &playThreadID));
    }

}




DWORD WINAPI DrawVideo(LPVOID params)
{
    /* This is the effect calculation thread launched from doit()
     * It jumps off (for the entire current play duration) into
     * one of the effect routines.
     */
    int i, count;
    char *dataPtr;
    DMint64 ust, oldmsc, newmsc;
    DMpv p[10];

    oldmsc = -1;
    ZC (wglMakeCurrent(_deviceContext, _renderContext));

    //Send the empty buufers for capturing.
    for (i = 0; i < _RVBUFS; i++) {
        p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
        p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)(_videoBufs + i*_videoSize);
        p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _videoSize; 
        p[_VC_UST_INDEX].param = DM_VIDEO_UST;
        p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
        p[_VC_BUFFER_PV_END].param = DM_END;

        DC (dmPathSendBuffers(_hVideoPath, p));
    }

    // Start video samples flowing along path. 
    DC (dmPathBeginTransfer(_hVideoPath) );

    while (_IamPlaying == 1) {
        do {    /* Get pointer to next valid field */ 
            checkReceiveQueue(&dataPtr, &ust, &newmsc);
        } while (dataPtr == NULL);

        if ((oldmsc != -1) && (newmsc != oldmsc + 1)) {
            fprintf(stderr, "MSC drop old = %lld, ", oldmsc);
            fprintf(stderr, "new = %lld\n", newmsc);
        }
        oldmsc = newmsc;

        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, _uSize, _vSize, _tex_format, 
                        GL_UNSIGNED_BYTE, dataPtr);
        glCallList(1);

        if (_dbuffer) {
            SwapBuffers(_deviceContext);
        }

        //Indicate te buffer is free to be used again.
        p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
        p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)(dataPtr);
        p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _videoSize; 
        p[_VC_UST_INDEX].param = DM_VIDEO_UST;
        p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
        p[_VC_BUFFER_PV_END].param = DM_END;

        DC (dmPathSendBuffers(_hVideoPath, p));
    }


    // Surrender the graphics context.
    glFlush();
    ZC (wglMakeCurrent(_deviceContext, NULL));
   
    // Shut down video 
    fprintf(stderr, "Ending video transfer\n");
    DC (dmPathEndTransfer(_hVideoPath) );
    DC (dmPathGetReceiveQueueMessageCount(_hVideoPath, &count)); 
    
    for (i = 0; i < count; i++) {
        checkReceiveQueue(&dataPtr, &ust, &newmsc);
    }

    --_IamPlaying;
    ExitThread (-1); 
}


void Quit(HWND wnd)
{
    int i;

    /* Stop play */
    if (_IamPlaying) doit();

	wglMakeCurrent(_deviceContext, _renderContext);
	wglDeleteContext(_renderContext);

	ReleaseDC(wnd, _deviceContext);

	PostQuitMessage(0);
}


void initVideo(dmJackTypeEnum jacktype)
{
    int i, numbufs;
    int bytes_per_pixel;
    DMpv *sysCaps, *jackCaps, *pathCaps;
    DMint64 *pJacks, *pPaths;
    DMint32 nJacks, nPaths;
    DMpv p[10];
    HDMJACK hJack;
   
    bytes_per_pixel  = (_packing == DM_FORMAT_CbYCrY_8_601) ? 2 : 4; 
    
    DC (dmSystemGetCapabilities(DM_SYSTEM_LOCALHOST, &sysCaps));
    pJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].value.pInt64;
    nJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].length;

    // Find a jack of the given type, i.e. composite, digial, etc.).
    for (i = 0; i < nJacks; i++) {
        DC (dmJackGetCapabilities(pJacks[i], &jackCaps));

        if (jackCaps[DM_JACK_DIRECTION_INDEX].value.int32 == DM_DIRECTION_IN &&
            jackCaps[DM_JACK_TYPE_INDEX].value.int32 == jacktype){
            break;
        }
    }

    ERR ((i == nJacks), "initVideo: can't find a jack");

    // Make sure there is a signal present on the desired jack.
    DC (dmJackOpen(pJacks[i], NULL, &hJack));
    i = 0;
    p[i].param = DM_JACK_SIGNAL_PRESENT_INT32; i++;
    p[i].param = DM_END;                       i++;
    DC (dmJackGetControls(hJack, p));
    ERR ((p[0].value.int32 == DM_TIMING_NONE), "No signal present for desired jack");
    ERR ((p[0].value.int32 == DM_TIMING_UNKNOWN), "Unknown signal present for desired jack");

    // For the given jack, find the dev to mem path.
    pPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].value.pInt64;
    nPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].length;
    for (i = 0; i < nPaths; i++) {
        DC (dmPathGetCapabilities(pPaths[i], &pathCaps));
        if (pathCaps[DM_PATH_TYPE_INDEX].value.int32 == DM_PATH_TYPE_DEV_TO_MEM) 
            break;
    }

    ERR ((i == nPaths), "initVideo: can't find a path");

    DC (dmPathOpen(pPaths[i], NULL, &_hVideoPath));

    //Set the image format and interleave mode on the path.
    i = 0;
    dmPvSetInt32Param( p[i], DM_IMAGE_FORMAT_INT32, _packing); i++;
    dmPvSetInt32Param( p[i], DM_IMAGE_INTERLEAVE_MODE_INT32, DM_INTERLEAVE_MODE_SINGLE_FIELD); i++;	
    dmPvSetInt32Param( p[i], DM_END, 0 ); i++;
    DC (dmPathSendControls(_hVideoPath, p));


    //Now read back the settings needed for the file header.
    i = 0;
    p[i].param = DM_VIDEO_WIDTH_INT32;     i++;
    p[i].param = DM_VIDEO_HEIGHT_F1_INT32; i++;
    p[i].param = DM_VIDEO_HEIGHT_F2_INT32; i++;
    p[i].param = DM_END;                   i++;
    DC (dmPathQueryControls(_hVideoPath, p));

    _uSize = p[0].value.int32;	//image width;
    _vSize = p[1].value.int32;	//image height;
    _videoSize =  bytes_per_pixel * _vSize * _uSize;

    _videoBufs =  (char *) VirtualAlloc(NULL,  _videoSize*_RVBUFS, MEM_COMMIT,PAGE_READWRITE);
    ERR ((_videoBufs == NULL), "Failed to allocate video buffers");

}



int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc)
{
    // Routnie to check status of buffers coming from video hardware.

    DMint32 message;
    DMpv *reply;
    DMstatus ret;
    int rval = 0;
 
    *data = NULL;
    *ust = -1;
    *msc = -1;

    ret = dmPathReceiveMessage(_hVideoPath, &message, &reply);

    if (ret != DM_STATUS_NO_ERROR) {     //Did not get error free result.

		if (ret != DM_STATUS_RECEIVE_QUEUE_EMPTY) {  //Did not get empty queue
	    	fprintf(stderr, "dmPathReceiveMessage returned value %x\n", ret);
		}
		return -1;  //Either error or queue empty.
    }

    switch (message) {
    case DM_MESSAGE_TRANSFER_COMPLETE:
		*data = (unsigned char *)reply[_VC_BUFFER_INDEX].value.pByte;
		*ust = reply[_VC_UST_INDEX].value.int64;
		*msc = reply[_VC_MSC_INDEX].value.int64;
		break;
    case DM_MESSAGE_MISSED_FIELD:
        //fprintf(stderr, "DM_MESSAGE_MISSED_FIELD\n");
		break;
    case DM_MESSAGE_TRANSFER_ABORTED:
		*data = (unsigned char *)reply[0].value.pByte;
        //fprintf(stderr, "DM_MESSAGE_TRANSFER_ABORTED\n");
		break;
	case DM_MESSAGE_CONTROLS_SUCCESSFUL:
        //fprintf(stderr, "DM_MESSAGE_CONTROLS_SUCCESSFUL\n");
		break;
    default:
        fprintf(stderr, "dmPathReceiveMessage returned message %x\n", ret);
		break;
    }

    return rval;
}
