/*
 * Copyright (c) 1999 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_u = Plays uncompressed video fields from disk to video out
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of reads for per readv call.
 *                                Default is 12, 8 and 16 are also good values.
 *     fileName                 - File created by vgrab_u
 *
 * The ReadsFile system call is used (along with direct I/O) for getting
 * the data from disk.  
 *
 * The optional flag "-iov" allows the user to specify the number of fields
 * to read for each ReadFile call.  Each  video field is padded out
 * to a 16k boundary (the default is 12 fields per read - smaller values like
 * 6 and 8 are also good and use less memory).
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After doing initialization, the routine doit() is entered.
 *      - Inside doit(), the program creates a read thread (see ReadIt) and
 *        a display thread (PutField()). 
 *	    - There is a queue of buffers between the reader and the display
 *		  threads.  There are _BUFS buffers in the queue, each one big enough to 
 *		  hold one field.  The buffers are pointed to the array of pointers 
 *        _readvBufs[].
 *		- The reader thread takes from the head of the queue (_readvBufs[_dmhead]).
 *	    - The display thread adds fields to the tail of the queue (_readvBufs[_dmtail]).
 *      - The main thread just hangs out. 
 *
 *    Read Thread: 
 *      - ReadIt() simply cycles through the file.  Per iteration it loads _iovcnt
 *	      buffers into the tail of _readvBufs starting with the buffer
 *	      pointed to by _readvBufs[_dmtail]. 
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	      by _readSize).
 *	   
 *    Display Thread:
 *		- The display thread (see PutField) continually checks 
 *	      for valid data in _readvBufs[_dmhead] and sends it to video-out. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <sgidmedia.h>
#include "..\common\handy.h"
#include "..\common\dmhandy.h"
#include "..\common\fileutils.h"
#include "..\common\filepacking.h"

// When we send buffers to dmPath via dmPathSendBuffers, use
// the following indices for the image buffer, UST, and MSC.
#define _VC_BUFFER_INDEX    0
#define _VC_UST_INDEX       1
#define _VC_MSC_INDEX       2
#define _VC_BUFFER_PV_END   3

#define MAX_IOV 16 
#define PRELOAD 2

/* All global variables in this file are preceded by an "_".  */

/* Misc. declarations.	*/
fileHeader_t _f;
int _iovcnt = 12;	/* Number of reads per readv call.	*/
int _CurField = 0;	/* Current field being processed.	*/

/* video buffer declarations  */
char    **_readvBufs;	/* Array of buffer pointers with valid data read from disk */
int _dmhead, _dmtail;	/* Head and tail pointers for _readvBufs queue.	*/
int          _BUFS;		/* Number of dm buffers in _outpool		*/ 
int          _mscStep;
HDMPATH      _hVideoPath;

/*Procedure declarations */
void doit();
void initVideo(dmJackTypeEnum);
DWORD WINAPI PutField(LPVOID);
DWORD WINAPI ReadIt(LPVOID);
int FilePackingToDMFormat(int packing);

main(int argc, char *argv[])
{
    int i,n;
    char *Fname;
    char *bigBuffer;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
	    else {
	        Fname = argv[i];
	    }
    }

    fileInit(Fname, &_f, TRUE);

    /* Cheat a liitle here, it makes the reading a lot simpler. */
    if (_f.Nfields < _iovcnt) {
		_iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }
	
    // Set up an arry of pointers to buffers used to hold the video
    // fields read from disk.
 
    _BUFS = 3*_iovcnt; 
    _readvBufs =  (char **) VirtualAlloc(NULL,  _BUFS*sizeof(char *), MEM_COMMIT,PAGE_READWRITE);

    bigBuffer  =  (char * ) VirtualAlloc(NULL,  _BUFS*_f.readSize, MEM_COMMIT,PAGE_READWRITE);
    ERR ((bigBuffer == NULL), "Failed to allocate reader buffers");

    for (i=0; i < _BUFS; i++) {
		_readvBufs[i] = bigBuffer + i*_f.readSize;
    }

    fprintf(stderr, "Playing  %d fields of %dx%d using a read vector of length %d\n", 
                     _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Padded read size (bytes) = %d \n", _f.readSize);

    /* Set up video hardware  */
    initVideo(DM_JACK_TYPE_COMPOSITE);

    doit();
}


void doit()
{
    /* sproc the read and display threads */
    DWORD readPID, putPID;
    HANDLE putThread, readThread;
    LPVOID params;

    _dmhead = _dmtail = 0; 

    NC(readThread = CreateThread (NULL, 0, ReadIt  , NULL, 0, &readPID));

    NC(putThread  = CreateThread (NULL, 0, PutField, NULL, 0, &putPID));
    while (1) Sleep(100);

}

DWORD WINAPI PutField(LPVOID params)
{
    /* This routine constany examines _readvBufs[_dmhead] for a valid 
     * buffer to send to video out.  After sending a buffer and being told
     * the system has used it (see routine checkReceiveQueue) it increments
     * _dmhead. 
     *
     * The ReadIt() thread is constantly reading fields off of disk and 
     * placing them at the end (i.e. _dmtail) of _readvBufs.  This thread
     * consumes from the front (i.e. _dmhead) of _readvBufs.
     */

    int i;
    int count = 0;
    int playHead;
    char *buf, *dataPtr;
    DMint64 oldmsc = -1;
    DMint64 newmsc;
    DMint64 ust;
    DMpv p[10];

    playHead = _dmhead;

    while (1) {

        // "playHead" points to the next buffer which can be
        // sent to video.  The reader thread looks at "_dmhead"
        // which is incremented after the buffer has been returned
        // by the video subsystem.

        if (playHead != _dmtail) {	// _readvBufs queue has data.

        	// Grab the head of _readvBufs and send to video. 
        	buf = _readvBufs[playHead];
        	++playHead;
        	if (playHead == _BUFS) playHead = 0;

        	// Send the buffer to video out.
			p[_VC_BUFFER_INDEX].param = DM_IMAGE_BUFFER_POINTER;
			p[_VC_BUFFER_INDEX].value.pByte = (DMbyte *)buf;
			p[_VC_BUFFER_INDEX].length = p[_VC_BUFFER_INDEX].maxLength = _f.readSize;
			p[_VC_UST_INDEX].param = DM_VIDEO_UST;
			p[_VC_MSC_INDEX].param = DM_VIDEO_MSC;
			p[_VC_BUFFER_PV_END].param = DM_END;
			DC (dmPathSendBuffers(_hVideoPath, p));
        	count++;
        }
        else { 					// _readvBufs queue is empty.
	    	Sleep(10);
	    }

        /* Don't start the transfer until we have some fields queued up.        */
        if (count == PRELOAD) {
            fprintf(stderr, "TRANSFER\n");
			DC (dmPathBeginTransfer(_hVideoPath) );
        }

        // Check and see if previous buffers are available for reuse.
        // Need to check each passs to avoid deadlock with ReadIt().
		checkReceiveQueue(&dataPtr, &ust, &newmsc);

        if (dataPtr != NULL) {

		    // Well then I just got a buffer back AND they must come back in the
            // order they were sent.  The next buffer to come back is _readvBufs[_dmhead].
            // Increment _dmhead so ReadIt() knows another buffer is available.
		    i = _dmhead + 1;
            if (i == _BUFS) i = 0;
            _dmhead = i;

            //Check and see if this buffer went out in time.
            if ((oldmsc != -1) && (newmsc != oldmsc + _mscStep)) {
                fprintf(stderr, "MSC drop old = %lld, ", oldmsc);
                fprintf(stderr, "new = %lld\n", newmsc);
            }
            oldmsc = newmsc;
        }
    } 
}

DWORD WINAPI ReadIt(LPVOID params)
{
    /* This routine reads _iovcnt fields from disk into buffers
     * starting at location _readvBufs[_dmtail].  It must check 
     * for full queue conditions with the display thread (PutField()).
     * _dmtail is incremented after a sucessful read.
     */

    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;
    LONG hi_off, lo_off;
    DWORDLONG offset;

    offset = _f.headerSize + (DWORDLONG)_CurField * _f.readSize;
    hi_off = (LONG)(offset>>32);
    lo_off = (LONG)(offset & 0x00000000ffffffff);
    SetFilePointer(_f.fd, lo_off, &hi_off, FILE_BEGIN);

    while (1) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Now make sure there is room in _readvBufs to store 
         * "fieldsToRead" buffers (which the consumer PutField() will later
         * display. 
         */
		do {
	    	full = 0;
	    	/* Check for the first full condition.  If the incremented
	     	 * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
	    	tmphead = _dmhead;
	    	tmptail = _dmtail + fieldsToRead;
            if (tmptail == _BUFS) tmptail = 0;
            full = (tmptail == tmphead); 

	        /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).  This works because _dmtail only gets
             * inc'd by _iovcnt AND the number of buffers (_BUFS) is a multiple
             * of _iovcnt.
             */
	    	full |= ((tmphead > _dmtail) && (tmphead < _dmtail + fieldsToRead));

	    	if (full) {
				/* fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail); */
				Sleep(10);
	    	}
		} while (full);

        /* Read the fields. Remember, _readvBufs is really one big contiguous
         * chunk of memory so we cam read all the fields with one ReadFile call.
         */
        tmptail = _dmtail;
        ZC ( ReadFile(_f.fd, _readvBufs[_dmtail], _f.readSize*fieldsToRead, &i, NULL));
        tmptail += fieldsToRead;
        if (tmptail == _BUFS) tmptail = 0;

        /* Inc. _dmtail, indicating more vaild data is in _readvBufs.	*/
        _dmtail = tmptail;
	
        /* Check for file wrap.     */
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
            _CurField = 0;
            SetFilePointer(_f.fd, _f.headerSize, NULL, FILE_BEGIN);
            fprintf(stderr, "Done pass through file\n");
        }
   } 
}

void initVideo(dmJackTypeEnum jacktype)
{
    int i, numbufs;
    int format, capType;
    int videoHeight;
    DMpv *sysCaps, *jackCaps, *pathCaps;
    DMint64 *pJacks, *pPaths;
    DMint32 nJacks, nPaths;
    DMpv p[10];

    DC (dmSystemGetCapabilities(DM_SYSTEM_LOCALHOST, &sysCaps));
    pJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].value.pInt64;
    nJacks = sysCaps[DM_SYSTEM_JACK_IDS_INDEX].length;

    // Find a jack of the given type, i.e. composite, digial, etc.).
    for (i = 0; i < nJacks; i++) {
        DC (dmJackGetCapabilities(pJacks[i], &jackCaps));

        if (jackCaps[DM_JACK_DIRECTION_INDEX].value.int32 == DM_DIRECTION_OUT &&
            jackCaps[DM_JACK_TYPE_INDEX].value.int32 == jacktype){
            break;
        }
    }

    ERR ((i == nJacks), "initVideo: can't find a jack");

    // For the given jack, find the dev to mem path.
    pPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].value.pInt64;
    nPaths = jackCaps[DM_JACK_PATH_IDS_INDEX].length;
    for (i = 0; i < nPaths; i++) {
        DC (dmPathGetCapabilities(pPaths[i], &pathCaps));
        if (pathCaps[DM_PATH_TYPE_INDEX].value.int32 == DM_PATH_TYPE_MEM_TO_DEV) break;
    }

    ERR ((i == nPaths), "initVideo: can't find a path");

    DC (dmPathOpen(pPaths[i], NULL, &_hVideoPath));

    format = FilePackingToDMFormat(_f.packing);

    if (_f.vSize < 400) {
        capType =  DM_INTERLEAVE_MODE_SINGLE_FIELD;
        videoHeight = _f.vSize;
        _mscStep = 1;
    }
    else {
        capType = DM_INTERLEAVE_MODE_INTERLEAVED;
        videoHeight = _f.vSize / 2;
        _mscStep = 2;
    }

    fprintf(stderr, "Video height = %d, sample height = %d msc step = %d\n", 
                     videoHeight, _f.vSize, _mscStep);

    //Set the image format and interleave mode on the path.
    i = 0;
    dmPvSetInt32Param(p[3], DM_VIDEO_WIDTH_INT32, _f.uSize);
    dmPvSetInt32Param(p[1], DM_VIDEO_HEIGHT_F1_INT32, videoHeight);
    dmPvSetInt32Param(p[2], DM_VIDEO_HEIGHT_F2_INT32, videoHeight);
    dmPvSetInt32Param(p[4], DM_IMAGE_FORMAT_INT32, format);
    dmPvSetInt32Param(p[0], DM_IMAGE_INTERLEAVE_MODE_INT32, capType);	
    dmPvSetInt32Param(p[5], DM_END, 0);
    DC (dmPathSendControls(_hVideoPath, p));

}

int checkReceiveQueue(unsigned char **data, DMint64 *ust, DMint64 *msc)
{
    // Routine to check for the next available buffer to load
    // with data.  Buffers are returned to the app. for reuse in the
    // order they are sent.

    DMint32 message;
    DMpv *reply;
    DMstatus ret;
    int rval = 0;
 
    *data = NULL;
    *ust = -1;
    *msc = -1;

    ret = dmPathReceiveMessage(_hVideoPath, &message, &reply);

    if (ret != DM_STATUS_NO_ERROR) {     //Did not get error free result.

		if (ret != DM_STATUS_RECEIVE_QUEUE_EMPTY) {  //Did not get empty queue
	    	fprintf(stderr, "dmPathReceiveMessage returned value %x\n", ret);
		}
		return -1;  //Either error or queue empty.
    }

    switch (message) {
    case DM_MESSAGE_TRANSFER_COMPLETE:
		*data = (unsigned char *)reply[_VC_BUFFER_INDEX].value.pByte;
		*ust = reply[_VC_UST_INDEX].value.int64;
		*msc = reply[_VC_MSC_INDEX].value.int64;
		break;
    case DM_MESSAGE_MISSED_FIELD:
		*data = NULL;
        //fprintf(stderr, "DM_MESSAGE_MISSED_FIELD\n");
		break;
    case DM_MESSAGE_TRANSFER_ABORTED:
		*data = (unsigned char *)reply[0].value.pByte;
        fprintf(stderr, "DM_MESSAGE_TRANSFER_ABORTED\n");
		break;
	case DM_MESSAGE_CONTROLS_SUCCESSFUL:
		*data = NULL;
        fprintf(stderr, "DM_MESSAGE_CONTROLS_SUCCESSFUL\n");
		break;
    default:
		*data = NULL;
        //fprintf(stderr, "dmPathReceiveMessage returned message %x\n", ret);
		break;
    }

    return rval;
}
